// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file UmfPackSolver.h */

#ifndef AMDIS_UMFPACKSOLVER_H
#define AMDIS_UMFPACKSOLVER_H

#if defined HAVE_UMFPACK && defined MTL_HAS_UMFPACK

#include <iostream>
#include <boost/numeric/mtl/operation/two_norm.hpp>
#include <boost/numeric/mtl/interface/umfpack_solve.hpp>
#include "MTL4Solver.h"

namespace AMDiS {

  /**
   * \ingroup Solver
   * 
   * \brief
   * Wrapper for the external UMFPACK solver:
   *   http://www.cise.ufl.edu/research/sparse/umfpack/
   *
   * This is a direct solver for large sparse matrices.
   */
  template< typename Matrix >
  struct Umfpack_runner 
  {
    typedef Matrix matrix_type ;
    OEMSolver& oem;
    Umfpack_runner(OEMSolver* oem_):
      oem(*oem_),
      solver(NULL),
      store_symbolic(0),
      symmetric_strategy(0),
      alloc_init(0.7)
    {
      FUNCNAME("Umfpack_runner::Umfpack_runner()");

      TEST_EXIT_DBG(oem_ != NULL)("Need real OEMSolver\n"); 
      Parameters::get(oem.getName() + "->store symbolic", store_symbolic);
      Parameters::get(oem.getName() + "->symmetric strategy", symmetric_strategy);
      Parameters::get(oem.getName() + "->alloc init", alloc_init);
    }

    void init(const Matrix& A)
    {
      FUNCNAME("Umfpack_runner::init()")

      if (solver != NULL) {
	delete solver;
	solver = NULL;
      }

      solver = new mtl::matrix::umfpack::solver<matrix_type>(A, symmetric_strategy, alloc_init);
    }

    template< typename Vector> 
    int solve(const Matrix& A, Vector& x, Vector& b) 
    { 
      FUNCNAME("Umfpack_runner::solve()");
      TEST_EXIT(solver != NULL)("The umfpack solver was not initialized\n");
#if 0 
      if (store_symbolic)
	solver->update_numeric();
      else
	solver->update();
#endif

      int code = (*solver)(x, b);
      if (oem.getInfo() > 0) {
        mtl::dense_vector<typename matrix_type::value_type> r(b); 
        r -= A * x; 
        double residual = two_norm(r);
        oem.setResidual(residual);
        oem.setErrorCode(code);
        std::cout << "UmfPackSolver: ||b-Ax|| = " << residual << "\n";
	TEST_EXIT(residual <= oem.getTolerance())("Tolerance tol = %e could not be reached!\n Set tolerance by '->solver->tolerance:' \n", oem.getTolerance());	
      }
      return code;
    }

    void exit()
    {}

    ~Umfpack_runner() 
    {
      if (solver != NULL) {	
	delete solver;
	solver == NULL;
      }
    }

  private:
    mtl::matrix::umfpack::solver<matrix_type> *solver;

    int store_symbolic;

    int symmetric_strategy;

    double alloc_init;
  };

  using namespace MTLTypes;
  class UmfPackSolver : public MTL4Solver<MTLMatrix, MTLVector, Umfpack_runner< MTLMatrix > >
  {
  protected:

  public:
    /// Creator class used in the OEMSolverMap.
    class Creator : public OEMSolverCreator
    {
    public:
      virtual ~Creator() {}
      
      /// Returns a new UmfPackSolver object.
      OEMSolver* create() 
      { 
	return new UmfPackSolver(this->name); 
      }
    };

    /// Constructor
    UmfPackSolver(std::string name) 
      : MTL4Solver< MTLMatrix, MTLVector, Umfpack_runner< MTLMatrix > >(name)
    {}
    
    int solveSystem(const SolverMatrix<Matrix<DOFMatrix*> >& A,
		    SystemVector& x, 
		    SystemVector& b,
		    VectorialMapper& m,
		    bool createMatrixData,
		    bool storeMatrixData) 
    {
      return solve(A, x, b, m, createMatrixData, storeMatrixData);
    }       
  };
  
}

#endif // HAVE_UMFPACK

#endif // AMDIS_UMFPACKSOLVER_H
