// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file PetscSolver.h */

#ifndef AMDIS_PETSC_SOLVER_H
#define AMDIS_PETSC_SOLVER_H

#include <set>
#include <map>
#include <mpi.h>
#include <petsc.h>
#include <petscsys.h>
#include <petscao.h>
#include <petscksp.h>

#include "AMDiS_fwd.h"
#include "Global.h"
#include "Initfile.h"
#include "DOFMatrix.h"
#include "parallel/MeshDistributor.h"
#include "parallel/ParallelCoarseSpaceMatVec.h"

namespace AMDiS {

  using namespace std;

  /**
   * Create an abstract interface to an arbitrary PETSc solver. This class is
   * based on \ref ParallelCoarseSpaceMatVec to support for solvers which make
   * use of a coarse grid problem.
   */
  class PetscSolver : public ParallelCoarseSpaceMatVec
  {
  public:
    PetscSolver();

    virtual ~PetscSolver() {}

    /** \brief
     * Create a PETSc matrix. The given DOF matrices are used to create the nnz 
     * structure of the PETSc matrix and the values are transfered to it.
     *
     * \param[in] mat
     */
    virtual void fillPetscMatrix(Matrix<DOFMatrix*> *mat) = 0;

    /** \brief
     * Create a PETSc vector and fills it with the rhs values of the system.
     *
     * \param[in] vec
     */
    virtual void fillPetscRhs(SystemVector *vec) = 0;

    /// Use PETSc to solve the linear system of equations
    virtual void solvePetscMatrix(SystemVector &vec, AdaptInfo *adaptInfo) = 0;

    virtual void solve(Vec &rhs, Vec &sol);

    virtual void solveGlobal(Vec &rhs, Vec &sol);

    /// Destroys all matrix data structures.
    virtual void destroyMatrixData() = 0;

    /// Detroys all vector data structures.
    virtual void destroyVectorData() = 0;

    virtual Flag getBoundaryDofRequirement()
    {
      return 0;
    }

    KSP getSolver() 
    { 
      return kspInterior; 
    }

    PC getPc() 
    { 
      return pcInterior; 
    }

    void setKspPrefix(std::string s)
    {
      kspPrefix = s;
    }

    void setRemoveRhsNullspace(bool b)
    {
      removeRhsNullspace = b;
    }

    /// Adds a new vector to the basis of the operator's nullspace.
    void addNullspaceVector(SystemVector *vec)
    {
      nullspace.push_back(vec);
    }

    /// Sets the nullspace to be constant for some specific components.
    void setConstantNullspace(vector<int> &components)
    {
      constNullspaceComponent = components;
    }

    /// Sets the nullspace to be constant for a specific component.
    void setConstantNullspace(int component)
    {
      constNullspaceComponent.clear();
      constNullspaceComponent.push_back(component);
    }

  protected:
    /** \brief
     * Copies between to PETSc vectors by using different index sets for the
     * origin and the destination vectors.
     *
     * \param[in]   originVec    The PETSc vector from which we copy from.
     * \param[out]  destVec      The PETSc vector we copy too.
     * \param[in]   originIndex  Set of global indices referring to the 
     *                           origin vector.
     * \param[in]   destIndex    Set of global indices referring to the
     *                           destination vector.
     */
    void copyVec(Vec& originVec, Vec& destVec, 
		 vector<int>& originIndex, vector<int>& destIndex);

  protected:
    /// PETSc solver object
    KSP kspInterior;

    /// PETSc preconditioner object
    PC pcInterior;

    /// A set of vectors that span the null space of the operator.
    vector<SystemVector*> nullspace;

    /// KSP database prefix
    string kspPrefix;

    /// If true, the constant null space is projected out of the RHS vector. It
    /// depends on the specific PETSc solver if it considers this value.
    bool removeRhsNullspace;

    bool hasConstantNullspace;

    vector<int> constNullspaceComponent;
  };


} // namespace AMDiS

#endif
