#include "AMDiS.h"
#include "NavierStokes_TaylorHood_RB.h"
#include "navierStokes.h"
#include "time/ExtendedRosenbrockAdaptInstationary.h"
#include "Refinement.h"
#include "MeshFunction_Level.h"

#include "boost/date_time/posix_time/posix_time.hpp"

using namespace AMDiS;
using namespace boost::posix_time;



class NS_Channel : public NavierStokes_TaylorHood_RB
{
public:
  NS_Channel(std::string name_) : NavierStokes_TaylorHood_RB(name_) {}

  void solveInitialProblem(AdaptInfo *adaptInfo)
  {
    super::solveInitialProblem(adaptInfo);

    size_t nVertices = 0;
    WorldVector<double> x;
    Parameters::get("obstacle->num vertices",nVertices);
    std::vector<WorldVector<double> > v(nVertices,x);
    for (size_t i = 0; i < nVertices; i++)
      Parameters::get("obstacle->vertex["+boost::lexical_cast<std::string>(i)+"]",v[i]);
    v.push_back(v[0]);

    SignedDistRefinement refFunction(getMesh());
    RefinementLevelCoords2 refinement(
      getFeSpace(),
      &refFunction,
      new Polygon(v));

    // initial refinement
    refinement.refine(10);
  }
protected:
  void fillBoundaryConditions()
  { FUNCNAME("NS_DrivenCavity::fillBoundaryConditions()");
    
    AbstractFunction<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);
    size_t dow = Global::getGeo(WORLD);

    // +------ 5 ------+
    // |               |
    // 2   # <--1      3
    // |               |
    // +------ 4 ------+


//     x[0] = 2.0;
//     x[1] = 2.0;
#if DEBUG != 0
  DOFVector<double>* polygon = new DOFVector<double>(getFeSpace(0), "polygon");
  polygon->interpol(new Polygon(v));
  VtkWriter::writeFile(polygon, "polygon.vtu");
#endif
    
    /// at rigid wall: no-slip boundary condition
    for (size_t i = 0; i < dow; i++) {
//       getProblem(0)->addImplicitDirichletBC(*(new Polygon(v)), i, i, *zero);
//       getProblem(0)->addSingularDirichletBC(x, i, i, *zero);
      getProblem(0)->addDirichletBC(1, i, i, zero);
      getProblem(0)->addDirichletBC(4, i, i, zero);
      getProblem(0)->addDirichletBC(5, i, i, zero);
    }

    double H = 4.1;
    double Um = 1.5;
    Parameters::get("mesh->H",H);
    Parameters::get("ns->Um",Um);

    /// at left wall: prescribed velocity
    getProblem(0)->addDirichletBC(2, 0, 0, new InflowBC(H,Um));
    getProblem(0)->addDirichletBC(2, 1, 1, zero);
  }
};


int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);

  NS_Channel nsProb("ns");
  nsProb.initialize(INIT_ALL | INIT_EXACT_SOLUTION);

  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", nsProb.getNumComponents());

  // adaption loop - solve ch-prob and ns-prob
//   AdaptInstationary adaptInstat("adapt", nsProb, adaptInfo, nsProb, adaptInfo);
  ExtendedRosenbrockAdaptInstationary<NS_Channel> adaptInstat("adapt", nsProb, adaptInfo, nsProb, adaptInfo);

  ptime start_time = microsec_clock::local_time();
  nsProb.initTimeInterface();
  int error_code = adaptInstat.adapt(); 
  time_duration td = microsec_clock::local_time()-start_time;

  MSG("elapsed time= %d sec\n", td.total_seconds());

  AMDiS::finalize();

  return error_code;
};
