#include "GeometryTools.h"

#define EPSILON FLT_EPSILON

namespace meshconv {

//                                      ###############
//                                      ##  General  ##
//                                      ###############

//----------------------------------------< solve_determinant4 >
double solve_determinant4(double x11, double x12, double x13, double x14,
                          double x21, double x22, double x23, double x24,
                          double x31, double x32, double x33, double x34,
                          double x41, double x42, double x43, double x44){
  return x11*(x22*x33*x44 + x23*x34*x42 + x24*x32*x43 - x24*x33*x42 - x23*x32*x44 - x22*x34*x43)
        -x12*(x21*x33*x44 + x23*x34*x41 + x24*x31*x43 - x24*x33*x41 - x23*x31*x44 - x21*x34*x43)
        +x13*(x21*x32*x44 + x22*x34*x41 + x24*x31*x42 - x24*x32*x41 - x22*x31*x44 - x21*x34*x42)
        -x14*(x21*x32*x43 + x22*x33*x41 + x23*x31*x42 - x23*x32*x41 - x22*x31*x43 - x21*x33*x42);
}

double solve_determinant3(double x11, double x12, double x13,
                          double x21, double x22, double x23,
                          double x31, double x32, double x33){
  return (x11*x22*x33 + x12*x23*x31 + x13*x21*x32)
	-(x13*x22*x31 + x12*x21*x33 + x11*x23*x32);
}

//                                      #################
//                                      ##  2D Worlds  ##
//                                      #################

//----------------------------------------< edge_length_2d >
double edge_length_2d(double lin0[], double lin1[]){
  long i;
  double p[2];

  for(i=0; i<2; ++i) p[i] = lin0[i] - lin1[i];
  return sqrt(p[0] * p[0] + p[1] * p[1]);
}

//----------------------------------------< boundingbox_triangle_2d >
void boundingbox_triangle_2d(double tri0[], double tri1[], double tri2[],
			     double min_corner[], double max_corner[]){
  min_corner[0] = std::min(std::min(tri0[0],tri1[0]),tri2[0]);
  min_corner[1] = std::min(std::min(tri0[1],tri1[1]),tri2[1]);
  max_corner[0] = std::max(std::max(tri0[0],tri1[0]),tri2[0]);
  max_corner[1] = std::max(std::max(tri0[1],tri1[1]),tri2[1]);
}

//----------------------------------------< centroid_of_box_2d >
void centroid_of_box_2d(double min_corner[], double max_corner[], double c[]){
  c[0] = 0.5*(max_corner[0]+min_corner[0]);
  c[1] = 0.5*(max_corner[1]+min_corner[1]);
}

//----------------------------------------< centroid_of_triangle_2d >
void centroid_of_triangle_2d(double tri0[], double tri1[], double tri2[], double c[]){
  c[0] = (tri0[0]+tri1[0],tri2[0])/3.0;
  c[1] = (tri0[1]+tri1[1],tri2[1])/3.0;
}

//----------------------------------------< point_in_triangle_2d >
bool point_in_triangle_2d(double p[], double a[], double b[], double c[]){
  long i;
  double v0[2], v1[2], v2[2];
  double dot00, dot01, dot02, dot11, dot12, invDenom, u, v;

  //compute direction vectors
  for(i=0; i<2; i++) v0[i] = c[i] - a[i];
  for(i=0; i<2; i++) v1[i] = b[i] - a[i];
  for(i=0; i<2; i++) v2[i] = p[i] - a[i];

  //compute dot products
  dot00 = v0[0]*v0[0] + v0[1]*v0[1]; //dot(v0, v0)
  dot01 = v0[0]*v1[0] + v0[1]*v1[1]; //dot(v0, v1)
  dot02 = v0[0]*v2[0] + v0[1]*v2[1]; //dot(v0, v2)
  dot11 = v1[0]*v1[0] + v1[1]*v1[1]; //dot(v1, v1)
  dot12 = v1[0]*v2[0] + v1[1]*v2[1]; //dot(v1, v2)

  //compute barycentric coordinates
  invDenom = 1.0 / (dot00 * dot11 - dot01 * dot01);
  u = (dot11 * dot02 - dot01 * dot12) * invDenom;
  v = (dot00 * dot12 - dot01 * dot02) * invDenom;

  //check if point is in triangle
  return (u >= 0.0) && (v >= 0.0) && (u + v <= 1.0);
}

//----------------------------------------< point_in_triangle_generous_2d >
bool point_in_triangle_generous_2d(double p[], double a[], double b[], double c[]){
  long i;
  double v0[2], v1[2], v2[2];
  double dot00, dot01, dot02, dot11, dot12, invDenom, u, v;

  //compute direction vectors
  for(i=0; i<2; i++) v0[i] = c[i] - a[i];
  for(i=0; i<2; i++) v1[i] = b[i] - a[i];
  for(i=0; i<2; i++) v2[i] = p[i] - a[i];

  //compute dot products
  dot00 = v0[0]*v0[0] + v0[1]*v0[1]; //dot(v0, v0)
  dot01 = v0[0]*v1[0] + v0[1]*v1[1]; //dot(v0, v1)
  dot02 = v0[0]*v2[0] + v0[1]*v2[1]; //dot(v0, v2)
  dot11 = v1[0]*v1[0] + v1[1]*v1[1]; //dot(v1, v1)
  dot12 = v1[0]*v2[0] + v1[1]*v2[1]; //dot(v1, v2)

  //compute barycentric coordinates
  invDenom = 1.0 / (dot00 * dot11 - dot01 * dot01);
  u = (dot11 * dot02 - dot01 * dot12) * invDenom;
  v = (dot00 * dot12 - dot01 * dot02) * invDenom;

  //check if point is in triangle
  return (u >= -EPSILON) && (v >= -EPSILON) && (u + v - 1.0 <= EPSILON);
}


//----------------------------------------< point_in_box_2d >
bool point_in_box_2d(double p[], double min_corner[], double max_corner[]){
  
  return (p[0] >= min_corner[0]) && (p[0] <= max_corner[0]) &&
	 (p[1] >= min_corner[1]) && (p[1] <= max_corner[1]);
}

//----------------------------------------< point_in_box_generous_2d >
bool point_in_box_generous_2d(double p[], double min_corner[], double max_corner[]){
  
  return (p[0]-min_corner[0] >= -EPSILON) && (p[0]-max_corner[0] <= EPSILON) &&
	 (p[1]-min_corner[1] >= -EPSILON) && (p[1]-max_corner[1] <= EPSILON);
}

//----------------------------------------< intersection_line_line_2d >
//tests whether the line segment defined by end points 'lin0' and 'lin1' intersects the connection
//between lamp and new_dof. If the line segment is intersected in one of its vertices the index of that
//vertex (0 or 1) is returned in hit_vertex otherwise hit_vertex is set to -1.
bool intersection_line_line_2d(double lin0[], double lin1[],
                               double lamp[], double new_dof[], long &hit_vertex){
  long i;
  double x=-1.0, y=-1.0, tmp;
  double p3[2], q3[2];
  hit_vertex = -1;

  for(i=0; i<2; ++i){
    p3[i] = lin1[i] - lin0[i];
    q3[i] = new_dof[i] - lamp[i];
  }

  //partial coincidence of small element and center-connection is a special case, it does not count as an
  //intersection
  tmp = p3[0]*q3[1] - p3[1]*q3[0];
  if(abs(tmp) > EPSILON){

    //no partial coincidence; test for intersection
    if(abs(q3[0]) < EPSILON){
      if(abs(p3[0]) > EPSILON){
        x = (lamp[0] - lin0[0]) / p3[0];
        y = (lamp[1] - lin0[1] - p3[1]*x) / -q3[1];
      }
    }
    else if(abs(q3[1]) < EPSILON){
      if(abs(p3[1]) > EPSILON){
        x = (lamp[1] - lin0[1]) / p3[1];
        y = (lamp[0] - lin0[0] - p3[0]*x) / -q3[0];
      }
    }
    else{
      if(abs(tmp) > EPSILON){
        x = ((lamp[0]-lin0[0])*q3[1] - (lamp[1]-lin0[1])*q3[0]) / tmp;
        y = (lamp[0] - lin0[0] - p3[0]*x) / -q3[0];
      }
    }

    if(x>-EPSILON && x-1<EPSILON && y>-EPSILON && y-1<-EPSILON){
      if(x<EPSILON) hit_vertex = 0;
      else if(x-1>-EPSILON) hit_vertex = 1;
      return true;
    }
  }

  return false;
}

//----------------------------------------< intersection_line_triangle_2d >
bool intersection_line_triangle_2d(double lin0[], double lin1[],
                                   double tri0[], double tri1[], double tri2[]){
  if(point_in_triangle_generous_2d(lin0, tri0, tri1, tri2)){
    return true;
  }
  if(point_in_triangle_generous_2d(lin1, tri0, tri1, tri2)){
    return true;
  }

  long hit_vertex;
  if(intersection_line_line_2d(tri0, tri1, lin0, lin1, hit_vertex)){
    return true;
  }
  if(intersection_line_line_2d(tri1, tri2, lin0, lin1, hit_vertex)){
    return true;
  }
  if(intersection_line_line_2d(tri2, tri0, lin0, lin1, hit_vertex)){
    return true;
  }

  return false;
}

//----------------------------------------< intersection_line_box_2d >
bool intersection_line_box_2d(double lin0[], double lin1[],
			      double min_corner[], double max_corner[]){
  double p0[2];
  double p1[2];
  p0[0] = min_corner[0]; p0[1] = max_corner[1];
  p1[0] = max_corner[0]; p1[1] = min_corner[1]; 
  
  long hit_vertex;
  if (intersection_line_line_2d(lin0,lin1,min_corner,p0,hit_vertex))
    return true;
  if (intersection_line_line_2d(lin0,lin1,min_corner,p1,hit_vertex))
    return true;
  if (intersection_line_line_2d(lin0,lin1,max_corner,p0,hit_vertex))
    return true;
  if (intersection_line_line_2d(lin0,lin1,max_corner,p1,hit_vertex))
    return true;
  
  return false;
}

//----------------------------------------< intersection_box_box_2d >
bool intersection_box_box_2d(double min_corner0[], double max_corner0[],
			     double min_corner1[], double max_corner1[]){
  double p[2];
  
  if (edge_length_2d(min_corner0,max_corner0) < edge_length_2d(min_corner1,max_corner1)) {
  // box0 is smaller than box1
    if (point_in_box_2d(min_corner0, min_corner1, max_corner1))
      return true;
    if (point_in_box_2d(max_corner0, min_corner1, max_corner1))
      return true;
    
    p[0] = min_corner0[0]; p[1] = max_corner0[1];
    if (point_in_box_2d(p, min_corner1, max_corner1))
      return true;
    
    p[0] = max_corner0[0]; p[1] = min_corner0[1];    
    if (point_in_box_2d(p, min_corner1, max_corner1))
      return true;
  } else {
  // box1 is smaller than box0
    if (point_in_box_2d(min_corner1, min_corner0, max_corner0))
      return true;
    if (point_in_box_2d(max_corner1, min_corner0, max_corner0))
      return true;
    
    p[0] = min_corner1[0]; p[1] = max_corner1[1];
    if (point_in_box_2d(p, min_corner0, max_corner0))
      return true;
    
    p[0] = max_corner1[0]; p[1] = min_corner1[1];    
    if (point_in_box_2d(p, min_corner0, max_corner0))
      return true;
  }

  return false;
}

//----------------------------------------< intersection_triangle_box_2d >
bool intersection_triangle_box_2d(double tri0[], double tri1[], double tri2[],
				  double min_corner[], double max_corner[]){
  // vertices of triangle covered by box
  if (point_in_box_generous_2d(tri0,min_corner,max_corner))
    return true;
  if (point_in_box_generous_2d(tri1,min_corner,max_corner))
    return true;
  if (point_in_box_generous_2d(tri2,min_corner,max_corner))
    return true;
  
  // bounding box of triangle does not intersect the box
  double p0[2];
  double p1[2];
  boundingbox_triangle_2d(tri0,tri1,tri2,p0,p1);
  if (!intersection_box_box_2d(min_corner,max_corner,p0,p1))
    return false;
  
  // vertices of box covered by triangle
  if (point_in_triangle_generous_2d(min_corner, tri0, tri1, tri2))
    return true;
  if (point_in_triangle_generous_2d(max_corner, tri0, tri1, tri2))
    return true;
  
  p0[0] = min_corner[0]; p0[1] = max_corner[1];
  if (point_in_triangle_generous_2d(p0, tri0, tri1, tri2))
    return true;
  
  p1[0] = max_corner[0]; p1[1] = min_corner[1]; 
  if (point_in_triangle_generous_2d(p1, tri0, tri1, tri2))
    return true;
  
  // edges of triangle intersect box
  if (intersection_line_box_2d(tri0,tri1,min_corner,max_corner))
    return true;
  if (intersection_line_box_2d(tri1,tri2,min_corner,max_corner))
    return true;
  if (intersection_line_box_2d(tri2,tri0,min_corner,max_corner))
    return true;
  
  return false;
}

//----------------------------------------< distance_point_line_2d >
//calculates the distance between a given point and the line segment between 'lin0' and 'lin1'.
double distance_point_line_2d(double point[], double lin0[], double lin1[]){
  double gx, gy, fx, fy, dx, dy, t;
  //the (complete) line through 'lin0' and 'lin1' is given by 'l = lin0 + t*g' with g as follows:
  gx = lin1[0] - lin0[0];
  gy = lin1[1] - lin0[1];

  //the normal-vector from 'point' to l is 'v = lin0 + t*g - point' for some t which is determined
  // by 'v*g = 0'
  t = (-gx * (lin0[0]-point[0]) - gy * (lin0[1]-point[1])) / (gx*gx + gy*gy);

  //if the orthogonal projection of 'point' onto l is not on the given line segment then one of the
  //vertices 'lin0' or 'lin1' is the nearest point to 'point'
  if(t<0){
    //'lin0' is nearest
    fx = lin0[0];
    fy = lin0[1];
  }
  else if(t>1){
    //'lin1' is nearest
    fx = lin1[0];
    fy = lin1[1];
  }
  else{
    //orthogonal projection is nearest
    fx = lin0[0] + t * gx;
    fy = lin0[1] + t * gy;
  }

  //calculate distance
  dx = point[0] - fx;
  dy = point[1] - fy;
  return sqrt(dx*dx + dy*dy);
}

//----------------------------------------< distance_point_line_with_intersection_2d >
//calculates the distance between a given point and the line-segment between 'lin0' and 'lin1'.
//This version also returns the intersection-point of the perpendicular with the given line-segment.
double distance_point_line_with_intersection_2d(double point[], double lin0[], double lin1[],
                                                                           double intersection[]){
  double gx, gy, dx, dy, t;
  //the (complete) line through 'lin0' and 'lin1' is given by 'l = lin0 + t*g' with g as follows:
  gx = lin1[0] - lin0[0];
  gy = lin1[1] - lin0[1];

  //the normal-vector from 'point' to l is 'v = lin0 + t*g - point' for some t which is determined
  // by 'v*g = 0'
  t = (-gx * (lin0[0]-point[0]) - gy * (lin0[1]-point[1])) / (gx*gx + gy*gy);

  //if the orthogonal projection of 'point' onto l is not on the given line segment then one of the
  //vertices 'lin0' or 'lin1' is the nearest point to 'point'
  if(t<0){
    //'lin0' is nearest
    intersection[0] = lin0[0];
    intersection[1] = lin0[1];
  }
  else if(t>1){
    //'lin1' is nearest
    intersection[0] = lin1[0];
    intersection[1] = lin1[1];
  }
  else{
    //orthogonal projection is nearest
    intersection[0] = lin0[0] + t * gx;
    intersection[1] = lin0[1] + t * gy;
  }

  //calculate distance
  dx = point[0] - intersection[0];
  dy = point[1] - intersection[1];
  return sqrt(dx*dx + dy*dy);
}


//----------------------------------------< distance_point_box_2d >
double distance_point_box_2d(double point[], double min_corner[], double max_corner[]){
  if (point_in_box_2d(point,min_corner,max_corner))
    return 0.0;
  
  double p[2];
  double dist = edge_length_2d(point, min_corner);  
  dist = std::min(dist, edge_length_2d(point, min_corner)); 
  p[0] = min_corner[0]; p[1] = max_corner[1];
  dist = std::min(dist, edge_length_2d(point, p));   
  p[0] = max_corner[0]; p[1] = min_corner[1];    
  dist = std::min(dist, edge_length_2d(point, p));
  
  return dist;
}

//----------------------------------------< distance_point_triangle_2d >
double distance_point_triangle_2d(double point[], double tri0[], double tri1[], double tri2[]){
  if (point_in_triangle_2d(point,tri0,tri1,tri2))
    return 0.0;
  
  double dist = edge_length_2d(point, tri0);  
  dist = std::min(dist, edge_length_2d(point, tri1)); 
  dist = std::min(dist, edge_length_2d(point, tri2));
  
  return dist;
}

bool point_in_polygon(double point[], const std::vector<AMDiS::WorldVector<double> > &vertices) {
  bool inside = false;

  int i, j;
  for (i = 0, j = vertices.size()-1; i < vertices.size(); j = i++) {
    if ((((vertices[i][1] <= point[1]) && (point[1] < vertices[j][1])) ||
         ((vertices[j][1] <= point[1]) && (point[1] < vertices[i][1]))) &&
        (point[0] < (vertices[j][0] - vertices[i][0]) * (point[1] - vertices[i][1]) / (vertices[j][1] - vertices[i][1]) + vertices[i][0]))
      inside = !inside;
  }

  return inside;
}

//----------------------------------------< triangle_area_2d >
double triangle_area_2d(double tri0[], double tri1[], double tri2[]){
  return 0.5 * abs((tri1[0] - tri0[0]) * (tri2[1] - tri0[1]) - (tri2[0] - tri0[0]) * (tri1[1] - tri0[1]));
}

//                                      #################
//                                      ##  3D Worlds  ##
//                                      #################

//----------------------------------------< points_identical_3d >
bool points_identical_3d(double p1[], double p2[]){
  long i;

  for(i=0; i<3; ++i){
    if(abs(p1[i] - p2[i]) > EPSILON) break;
  }
  if(i == 3){
    return true;
  }else{
    return false;
  }
}

//----------------------------------------< normal_vector_3d >
void normal_vector_3d(double tri0[], double tri1[], double tri2[], double normal[]){
  long i;
  double p[3], q[3];

  for(i=0; i<3; ++i) p[i] = tri0[i] - tri1[i];
  for(i=0; i<3; ++i) q[i] = tri0[i] - tri2[i];
  normal[0] = p[1] * q[2] - p[2] * q[1];
  normal[1] = p[2] * q[0] - p[0] * q[2];
  normal[2] = p[0] * q[1] - p[1] * q[0];
}

//----------------------------------------< triangle_area_3d >
double triangle_area_3d(double tri0[], double tri1[], double tri2[]){
  long i;
  double p[3], q[3], n[3];

  for(i=0; i<3; ++i) p[i] = tri0[i] - tri1[i];
  for(i=0; i<3; ++i) q[i] = tri0[i] - tri2[i];
  n[0] = p[1] * q[2] - p[2] * q[1];
  n[1] = p[2] * q[0] - p[0] * q[2];
  n[2] = p[0] * q[1] - p[1] * q[0];
  return sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]) / 2;
}

//----------------------------------------< edge_length_3d >
double edge_length_3d(double lin0[], double lin1[]){
  long i;
  double p[3];

  for(i=0; i<3; ++i) p[i] = lin0[i] - lin1[i];
  return sqrt(p[0] * p[0] + p[1] * p[1] + p[2] * p[2]);
}

//----------------------------------------< triangle_max_edge_length_3d >
double triangle_max_edge_length_3d(double tri0[], double tri1[], double tri2[]){
  long i;
  double p[3], max_el, a;

  for(i=0; i<3; ++i) p[i] = tri0[i] - tri1[i];
  a = p[0] * p[0] + p[1] * p[1] + p[2] * p[2];
  max_el = a;

  for(i=0; i<3; ++i) p[i] = tri1[i] - tri2[i];
  a = p[0] * p[0] + p[1] * p[1] + p[2] * p[2];
  if(a > max_el) max_el = a;

  for(i=0; i<3; ++i) p[i] = tri2[i] - tri0[i];
  a = p[0] * p[0] + p[1] * p[1] + p[2] * p[2];
  if(a > max_el) max_el = a;

  return sqrt(max_el);
}

//----------------------------------------< unit_normal_vector_3d >
void unit_normal_vector_3d(double tri0[], double tri1[], double tri2[], double normal[]){
  long i;
  double b;
  double p[3], q[3];

  for(i=0; i<3; ++i) p[i] = tri0[i] - tri1[i];
  for(i=0; i<3; ++i) q[i] = tri0[i] - tri2[i];
  normal[0] = p[1] * q[2] - p[2] * q[1];
  normal[1] = p[2] * q[0] - p[0] * q[2];
  normal[2] = p[0] * q[1] - p[1] * q[0];
  b = sqrt(normal[0] * normal[0] + normal[1] * normal[1] + normal[2] * normal[2]);
  for(i=0; i<3; ++i) normal[i] /= b;
}

//----------------------------------------< degenerate_triangle_3d >
bool degenerate_triangle_3d(double tri0[], double tri1[], double tri2[]){
  long i;
  double p[3], q[3], n[3];

  for(i=0; i<3; ++i) p[i] = tri0[i] - tri1[i];
  for(i=0; i<3; ++i) q[i] = tri0[i] - tri2[i];
  n[0] = p[1] * q[2] - p[2] * q[1];
  n[1] = p[2] * q[0] - p[0] * q[2];
  n[2] = p[0] * q[1] - p[1] * q[0];
  if(abs(n[0])<EPSILON && abs(n[1])<EPSILON && abs(n[2])<EPSILON){
    return true;
  }else{
    return false;
  }
}

//----------------------------------------< boundingbox_triangle_3d >
void boundingbox_triangle_3d(double tri0[], double tri1[], double tri2[],
			     double min_corner[], double max_corner[]){
  min_corner[0] = std::min(std::min(tri0[0],tri1[0]),tri2[0]);
  min_corner[1] = std::min(std::min(tri0[1],tri1[1]),tri2[1]);
  min_corner[2] = std::min(std::min(tri0[2],tri1[2]),tri2[2]);
  max_corner[0] = std::max(std::max(tri0[0],tri1[0]),tri2[0]);
  max_corner[1] = std::max(std::max(tri0[1],tri1[1]),tri2[1]);
  max_corner[2] = std::max(std::max(tri0[2],tri1[2]),tri2[2]);
}

//----------------------------------------< centroid_of_box_3d >
void centroid_of_box_3d(double min_corner[], double max_corner[], double c[]){
  c[0] = 0.5*(max_corner[0]+min_corner[0]);
  c[1] = 0.5*(max_corner[1]+min_corner[1]);
  c[2] = 0.5*(max_corner[2]+min_corner[2]);
}

//----------------------------------------< centroid_of_triangle_3d >
void centroid_of_triangle_3d(double tri0[], double tri1[], double tri2[], double c[]){
  c[0] = (tri0[0]+tri1[0],tri2[0])/3.0;
  c[1] = (tri0[1]+tri1[1],tri2[1])/3.0;
  c[2] = (tri0[2]+tri1[2],tri2[2])/3.0;
}

//----------------------------------------< point_in_box_3d >
bool point_in_box_3d(double p[], double min_corner[], double max_corner[]){
  
  return (p[0] >= min_corner[0]) && (p[0] <= max_corner[0]) &&
	 (p[1] >= min_corner[1]) && (p[1] <= max_corner[1]) &&
	 (p[2] >= min_corner[2]) && (p[2] <= max_corner[2]);
}

//----------------------------------------< point_in_box_generous_3d >
bool point_in_box_generous_3d(double p[], double min_corner[], double max_corner[]){
  
  return (p[0]-min_corner[0] >= -EPSILON) && (p[0]-max_corner[0] <= EPSILON) &&
	 (p[1]-min_corner[1] >= -EPSILON) && (p[1]-max_corner[1] <= EPSILON) &&
	 (p[2]-min_corner[2] >= -EPSILON) && (p[2]-max_corner[2] <= EPSILON);
}

//----------------------------------------< point_in_tetrahedron_3d >
bool point_in_tetrahedron_3d(double p[], double a[], double b[], double c[], double d[]){
  double d0, d1, d2, d3, d4;

  d0 = solve_determinant4(a[0], a[1], a[2], 1.0, b[0], b[1], b[2], 1.0,
                          c[0], c[1], c[2], 1.0, d[0], d[1], d[2], 1.0);
  d1 = solve_determinant4(p[0], p[1], p[2], 1.0, b[0], b[1], b[2], 1.0,
                          c[0], c[1], c[2], 1.0, d[0], d[1], d[2], 1.0);
  d2 = solve_determinant4(a[0], a[1], a[2], 1.0, p[0], p[1], p[2], 1.0,
                          c[0], c[1], c[2], 1.0, d[0], d[1], d[2], 1.0);
  d3 = solve_determinant4(a[0], a[1], a[2], 1.0, b[0], b[1], b[2], 1.0,
                          p[0], p[1], p[2], 1.0, d[0], d[1], d[2], 1.0);
  d4 = solve_determinant4(a[0], a[1], a[2], 1.0, b[0], b[1], b[2], 1.0,
                          c[0], c[1], c[2], 1.0, p[0], p[1], p[2], 1.0);
  //if(d0>=0.0){
  //  return (d1>=0.0 && d2>=0.0 && d3>=0.0 && d4>=0.0);
  //}else{
  //  return (d1<=0.0 && d2<=0.0 && d3<=0.0 && d4<=0.0);
  //}
  //the following code is highly experimental and should not be trusted!
  if(d0>=0.0){
    return (d1>=-EPSILON && d2>=-EPSILON && d3>=-EPSILON && d4>=-EPSILON);
  }else{
    return (d1<=EPSILON && d2<=EPSILON && d3<=EPSILON && d4<=EPSILON);
  }
}

//----------------------------------------< intersection_line_line_3d >
bool intersection_line_line_3d(double p1[], double p2[], double p3[], double p4[]){
  double p13[3], p43[3], p21[3];
  double d1343, d4321, d1321, d4343, d2121;
  double numer, denom;
  double mua, mub;

  p13[0] = p1[0] - p3[0];
  p13[1] = p1[1] - p3[1];
  p13[2] = p1[2] - p3[2];
  p43[0] = p4[0] - p3[0];
  p43[1] = p4[1] - p3[1];
  p43[2] = p4[2] - p3[2];
  p21[0] = p2[0] - p1[0];
  p21[1] = p2[1] - p1[1];
  p21[2] = p2[2] - p1[2];

  //test for collinearity
  if( abs(p21[1]*p43[2] - p21[2]*p43[1]) < EPSILON
      && abs(p21[2]*p43[0] - p21[0]*p43[2]) < EPSILON
      && abs(p21[0]*p43[1] - p21[1]*p43[0]) < EPSILON ){
    //lines are collinear; now test for an intersection
    double x11, x12, x21, x22, y11, y12, y21, y22, z11, z12, z21, z22;
    if(p1[0]<=p2[0]){
      x11 = p1[0];
      x12 = p2[0];
    }
    else{
      x11 = p2[0];
      x12 = p1[0];
    }
    if(p1[1]<=p2[1]){
      y11 = p1[1];
      y12 = p2[1];
    }
    else{
      y11 = p2[1];
      y12 = p1[1];
    }
    if(p1[2]<=p2[2]){
      z11 = p1[2];
      z12 = p2[2];
    }
    else{
      z11 = p2[2];
      z12 = p1[2];
    }
    if(p3[0]<=p4[0]){
      x21 = p3[0];
      x22 = p4[0];
    }
    else{
      x21 = p4[0];
      x22 = p3[0];
    }
    if(p3[1]<=p4[1]){
      y21 = p3[1];
      y22 = p4[1];
    }
    else{
      y21 = p4[1];
      y22 = p3[1];
    }
    if(p3[2]<=p4[2]){
      z21 = p3[2];
      z22 = p4[2];
    }
    else{
      z21 = p4[2];
      z22 = p3[2];
    }
    if(x11<=x22 && x12>=x21 && y11<=y22 && y12>=y21 && z11<=z22 && z12>=z21){
      if(abs(p21[0]) > EPSILON){
        mua = -p13[0]/p21[0];
        if(abs(p1[1] + p21[1]*mua - p3[1]) < EPSILON && abs(p1[2] + p21[2]*mua - p3[2]) < EPSILON){
          return true;
        }
      }else if(abs(p21[1]) > EPSILON){
        mua = -p13[1]/p21[1];
        if(abs(p1[0] + p21[0]*mua - p3[0]) < EPSILON && abs(p1[2] + p21[2]*mua - p3[2]) < EPSILON){
          return true;
        }
      }else{
        mua = -p13[2]/p21[2];
        if(abs(p1[0] + p21[0]*mua - p3[0]) < EPSILON && abs(p1[1] + p21[1]*mua - p3[1]) < EPSILON){
          return true;
        }
      }
    }
    return false;
  }

  d1343 = p13[0] * p43[0] + p13[1] * p43[1] + p13[2] * p43[2];
  d4321 = p43[0] * p21[0] + p43[1] * p21[1] + p43[2] * p21[2];
  d1321 = p13[0] * p21[0] + p13[1] * p21[1] + p13[2] * p21[2];
  d4343 = p43[0] * p43[0] + p43[1] * p43[1] + p43[2] * p43[2];
  d2121 = p21[0] * p21[0] + p21[1] * p21[1] + p21[2] * p21[2];

  denom = d2121 * d4343 - d4321 * d4321;
  if(abs(denom) < EPSILON){
    return false;
  }
  numer = d1343 * d4321 - d1321 * d4343;

  mua = numer / denom;
  mub = (d1343 + d4321 * mua) / d4343;

  if(mua<-EPSILON || mua-1>EPSILON || mub<-EPSILON || mub-1>EPSILON){
    return false;
  }
  if(abs(p1[0] + mua * p21[0] - p3[0] - mub * p43[0]) < EPSILON
         && abs(p1[1] + mua * p21[1] - p3[1] - mub * p43[1]) < EPSILON
         && abs(p1[2] + mua * p21[2] - p3[2] - mub * p43[2]) < EPSILON){
    return true;
  }else{
    return false;
  }
}

//----------------------------------------< intersection_line_line_with_intersection_3d >
bool intersection_line_line_with_intersection_3d(double p1[], double p2[], double p3[], double p4[],
                                                                               double intersection[]){
  double p13[3], p43[3], p21[3];
  double d1343, d4321, d1321, d4343, d2121;
  double numer, denom;
  double mua, mub;
  int ci = -1;

  p13[0] = p1[0] - p3[0];
  p13[1] = p1[1] - p3[1];
  p13[2] = p1[2] - p3[2];
  p43[0] = p4[0] - p3[0];
  p43[1] = p4[1] - p3[1];
  p43[2] = p4[2] - p3[2];
  p21[0] = p2[0] - p1[0];
  p21[1] = p2[1] - p1[1];
  p21[2] = p2[2] - p1[2];

  //test for collinearity
  if( abs(p21[1]*p43[2] - p21[2]*p43[1]) < EPSILON
      && abs(p21[2]*p43[0] - p21[0]*p43[2]) < EPSILON
      && abs(p21[0]*p43[1] - p21[1]*p43[0]) < EPSILON ){
    //lines are collinear; now test for an intersection
    double x11, x12, x21, x22, y11, y12, y21, y22, z11, z12, z21, z22;
    if(p1[0]<=p2[0]){
      x11 = p1[0];
      x12 = p2[0];
    }
    else{
      x11 = p2[0];
      x12 = p1[0];
    }
    if(p1[1]<=p2[1]){
      y11 = p1[1];
      y12 = p2[1];
    }
    else{
      y11 = p2[1];
      y12 = p1[1];
    }
    if(p1[2]<=p2[2]){
      z11 = p1[2];
      z12 = p2[2];
    }
    else{
      z11 = p2[2];
      z12 = p1[2];
    }
    if(p3[0]<=p4[0]){
      x21 = p3[0];
      x22 = p4[0];
    }
    else{
      x21 = p4[0];
      x22 = p3[0];
    }
    if(p3[1]<=p4[1]){
      y21 = p3[1];
      y22 = p4[1];
    }
    else{
      y21 = p4[1];
      y22 = p3[1];
    }
    if(p3[2]<=p4[2]){
      z21 = p3[2];
      z22 = p4[2];
    }
    else{
      z21 = p4[2];
      z22 = p3[2];
    }
    if(x11<=x22 && x12>=x21 && y11<=y22 && y12>=y21 && z11<=z22 && z12>=z21){
      if(abs(p21[0]) > EPSILON){
        mua = -p13[0] / p21[0];
        if(abs(p1[1] + p21[1]*mua - p3[1]) < EPSILON && abs(p1[2] + p21[2]*mua - p3[2]) < EPSILON) ci = 0;
      }else if(abs(p21[1]) > EPSILON){
        mua = -p13[1] / p21[1];
        if(abs(p1[0] + p21[0]*mua - p3[0]) < EPSILON && abs(p1[2] + p21[2]*mua - p3[2]) < EPSILON) ci = 1;
      }else{
        mua = -p13[2] / p21[2];
        if(abs(p1[0] + p21[0]*mua - p3[0]) < EPSILON && abs(p1[1] + p21[1]*mua - p3[1]) < EPSILON) ci = 2;
      }
      if(ci != -1){
        //determine best intersection-point
        //priority is p3(*) > p4 > p3 > p2 (one of these three points must be part of both line segments
        //(*): p3 is only priorized over p4 if it is not a vertex of the first line segment
        if(mua > EPSILON && mua-1< -EPSILON){
          intersection[0] = p3[0];
          intersection[1] = p3[1];
          intersection[2] = p3[2];
        }else{
          mub = (p4[ci]-p1[ci]) / p21[ci];
          if(mub > -EPSILON && mub-1 < EPSILON){
            intersection[0] = p4[0];
            intersection[1] = p4[1];
            intersection[2] = p4[2];
          }else if(mua > -EPSILON && mua-1 < EPSILON){
            intersection[0] = p3[0];
            intersection[1] = p3[1];
            intersection[2] = p3[2];
          }else{
            intersection[0] = p2[0];
            intersection[1] = p2[1];
            intersection[2] = p2[2];
          }
        }
        return true;
      }
    }
    return false;
  }

  d1343 = p13[0] * p43[0] + p13[1] * p43[1] + p13[2] * p43[2];
  d4321 = p43[0] * p21[0] + p43[1] * p21[1] + p43[2] * p21[2];
  d1321 = p13[0] * p21[0] + p13[1] * p21[1] + p13[2] * p21[2];
  d4343 = p43[0] * p43[0] + p43[1] * p43[1] + p43[2] * p43[2];
  d2121 = p21[0] * p21[0] + p21[1] * p21[1] + p21[2] * p21[2];

  denom = d2121 * d4343 - d4321 * d4321;
  if(abs(denom) < EPSILON){
    return false;
  }
  numer = d1343 * d4321 - d1321 * d4343;

  mua = numer / denom;
  mub = (d1343 + d4321 * mua) / d4343;
  if(mua<-EPSILON || mua-1>EPSILON || mub<-EPSILON || mub-1>EPSILON){
    return false;
  }

  //determine potential intersection-point
  intersection[0] = p1[0] + mua * p21[0];
  intersection[1] = p1[1] + mua * p21[1];
  intersection[2] = p1[2] + mua * p21[2];

  //check if this really is an intersection
  if(abs(intersection[0] - p3[0] - mub * p43[0]) < EPSILON
         && abs(intersection[1] - p3[1] - mub * p43[1]) < EPSILON
         && abs(intersection[2] - p3[2] - mub * p43[2]) < EPSILON){
    return true;
  }else{
    return false;
  }
}

//----------------------------------------< intersection_line_triangle_3d_chirkov >
//ATTENTION: function does not work for lines almost in the plane of the triangle!
//calculates the intersection of a line segment and a triangle if it exists. The intersection-point is not
//calculated, the last argument "sol" remains unchanged.
//    input:  a line segment (lin0, lin1), and a triangle (tri0, tri1, tri2)
//    return: 0      disjoint
//            1      intersection inside triangle
//            2,3,4  intersection in edge of triangle (edge_index is 0, 1, 2 respectively)
//this algorithm is based on an implementation by Nick Chirkov (journal of graphics tools 10(3):13-18, 2005).
int intersection_line_triangle_3d(double tri0[], double tri1[], double tri2[],
                                            double lin0[], double lin1[], double sol[]){
  long i, type;
  double org[3];
  double end[3];
  double dir[3];
  double x, y, z, d;

  //cube-collision-detection to determine necessity of real intersection check
  //note: this does not seem to improve performance unfortunately.
  //double p3[3],q3[3],p4[3],q4[3];
  //for(i=0; i<3; ++i){
  //  if(tri0[i] <= tri1[i]){
  //    if(tri0[i] <= tri2[i]){
  //      p3[i] = tri0[i];
  //      q3[i] = ((tri1[i]>=tri2[i])? tri1[i] : tri2[i]);
  //    }
  //    else{
  //      p3[i] = tri2[i];
  //      q3[i] = tri1[i];
  //    }
  //  }
  //  else{
  //    if(tri1[i] <= tri2[i]){
  //      p3[i] = tri1[i];
  //      q3[i] = ((tri0[i]>=tri2[i])? tri2[i] : tri0[i]);
  //    }
  //    else{
  //      p3[i] = tri2[i];
  //      q3[i] = tri0[i];
  //    }
  //  }
  //  if(lin0[i] <= lin1[i]){
  //    p4[i] = lin0[i];
  //    q4[i] = lin1[i];
  //  }
  //  else{
  //    p4[i] = lin1[i];
  //    q4[i] = lin0[i];
  //  }
  //}
  //if(p3[0]>q4[0] || q3[0]<p4[0] || p3[1]>q4[1] || q3[1]<p4[1] || p3[2]>q4[2] || q3[2]<p4[2]){
  //  return 0;
  //}

  double ax = tri1[0] - tri0[0];
  double ay = tri1[1] - tri0[1];
  double az = tri1[2] - tri0[2];
  double bx = tri2[0] - tri0[0];
  double by = tri2[1] - tri0[1];
  double bz = tri2[2] - tri0[2];
  x = ay * bz - az * by;
  y = az * bx - ax * bz;
  z = ax * by - ay * bx;
  d = x*tri0[0] + y*tri0[1] + z*tri0[2];

  static const double SQ = sqrt(1.0/3.0);
  double len = (x*x + y*y + z*z)*SQ;

  if(fabs(x)>len) type = 1;
  else if(fabs(y)>len) type = 2;
  else type = 3;

  for(i=0;i<=2;i++){
    org[i] = lin0[i]; //todo: org can be replaced by lin0 throughout the whole function
    end[i] = lin1[i]; //todo: replace end by lin1
    dir[i] = end[i] - org[i];
  }

  //determine intersection
  double signSrc = x*org[0] + y*org[1] + z*org[2] - d;
  double signDst = x*end[0] + y*end[1] + z*end[2] - d;
  if(signSrc*signDst > 0.0){
    return 0;
  }

  double di = signSrc - signDst;

  if(type == 1){
    double basey = org[1] - tri0[1];
    double basez = org[2] - tri0[2];
    double adelx = signSrc*(ay * dir[2] - az * dir[1]);
    if((adelx + di * (ay*basez-az*basey))
              * (signSrc*(dir[1]*bz-dir[2]*by) + di * (basey*bz-basez*by)) > 0.0){
      double e2y = tri1[1] - tri2[1];
      double e2z = tri1[2] - tri2[2];
      basey = org[1] - tri1[1];
      basez = org[2] - tri1[2];
      if((adelx + di * (ay*basez-az*basey))
              * (signSrc*(dir[1]*e2z-dir[2]*e2y) + di * (basey*e2z-basez*e2y)) > 0.0){
        goto test_edges;
      }
    }
  }
  else if(type == 2){
    double basex = org[0] - tri0[0];
    double basez = org[2] - tri0[2];
    double adely = signSrc*(az * dir[0] - ax * dir[2]);
    if((adely + di * (az*basex-ax*basez))
                 * (signSrc*(dir[2]*bx-dir[0]*bz) + di * (basez*bx-basex*bz)) > 0.0){
      double e2x = tri1[0] - tri2[0];
      double e2z = tri1[2] - tri2[2];
      basex = org[0] - tri1[0];
      basez = org[2] - tri1[2];
      if((adely + di * (az*basex-ax*basez))
                 * (signSrc*(dir[2]*e2x-dir[0]*e2z) + di * (basez*e2x-basex*e2z)) > 0.0){
        goto test_edges;
      }
    }
  }
  else{
    double basex = org[0] - tri0[0];
    double basey = org[1] - tri0[1];
    double adelz = signSrc*(ax * dir[1] - ay * dir[0]);
    if((adelz + di * (ax*basey-ay*basex))
                   * (signSrc*(dir[0]*by-dir[1]*bx) + di * (basex*by-basey*bx)) > 0.0){
      double e2x = tri1[0] - tri2[0];
      double e2y = tri1[1] - tri2[1];
      basex = org[0] - tri1[0];
      basey = org[1] - tri1[1];
      if((adelz + di * (ax*basey-ay*basex))
                   * (signSrc*(dir[0]*e2y-dir[1]*e2x) + di * (basex*e2y-basey*e2x)) > 0.0){
        goto test_edges;
      }
    }
  }

  return 0;

  test_edges:;
  if(intersection_line_line_3d(tri0, tri1, lin0, lin1)){
    return 4;
  }else if(intersection_line_line_3d(tri1, tri2, lin0, lin1)){
    return 2;
  }else if(intersection_line_line_3d(tri2, tri0, lin0, lin1)){
    return 3;
  }
  return 1;
}

//----------------------------------------< intersection_line_triangle_3d_softsurfer >
// Copyright 2001, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.

//calculates the intersection point of a line segment and a triangle if it exists
//    input:  a line segment (lin0, lin1), and a triangle (tri0, tri1, tri2)
//    output: sol = intersection point (if it exists)
//    return: 0      disjoint
//            1      intersection inside triangle
//            2,3,4  intersection in edge of triangle (edge_index is 0, 1, 2 respectively)
int intersection_line_triangle_3d_softsurfer(double tri0[], double tri1[], double tri2[],
                                    double lin0[], double lin1[], double sol[]){
  double u[3], v[3], n[3];    //triangle vectors
  double dir[3], w0[3], w[3]; //ray vectors
  double r, a, b;             //params to calc ray-plane intersect
  long i, e;
  double EPSILON2 = 100 * EPSILON;

  double uu, uv, vv, wu, wv, D;
  double s, t;
  double distances[3];

  //get triangle edge vectors and plane normal
  for(i=0; i<3; i++) u[i] = tri1[i] - tri0[i];
  for(i=0; i<3; i++) v[i] = tri2[i] - tri0[i];
  n[0] = u[1]*v[2] - u[2]*v[1];
  n[1] = u[2]*v[0] - u[0]*v[2];
  n[2] = u[0]*v[1] - u[1]*v[0];
  if(abs(n[0])<EPSILON && abs(n[1])<EPSILON && abs(n[2])<EPSILON){
    if(intersection_line_line_with_intersection_3d(tri0, tri1, lin0, lin1, sol)){
      return 4;
    }else if(intersection_line_line_with_intersection_3d(tri1, tri2, lin0, lin1, sol)){
      return 2;
    }else if(intersection_line_line_with_intersection_3d(tri2, tri0, lin0, lin1, sol)){
      return 3;
    }
  }

  for(i=0; i<3; i++) dir[i] = lin1[i] - lin0[i]; //line direction vector
  for(i=0; i<3; i++) w0[i] = lin0[i] - tri0[i];
  a = -(n[0] * w0[0] + n[1] * w0[1] + n[2] * w0[2]);
  b = n[0] * dir[0] + n[1] * dir[1] + n[2] * dir[2];
  if(abs(b) < EPSILON){ //ray is parallel to triangle plane
    if(abs(a) < EPSILON){
      //test whether one of the line's vertices lies inside the triangle
      for(e=0; e<2; e++){
        if(e==0) for(i=0; i<3; i++) sol[i] = lin0[i];
        else for(i=0; i<3; i++) sol[i] = lin1[i];

        uu = u[0] * u[0] + u[1] * u[1] + u[2] * u[2];
        uv = u[0] * v[0] + u[1] * v[1] + u[2] * v[2];
        vv = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
        for(i=0; i<3; i++) w[i] = sol[i] - tri0[i];

        wu = w[0] * u[0] + w[1] * u[1] + w[2] * u[2];
        wv = w[0] * v[0] + w[1] * v[1] + w[2] * v[2];
        D = uv * uv - uu * vv;

        s = (uv * wv - vv * wu) / D;
        if (s < -EPSILON || s-1.0 > EPSILON) continue; //sol is outside triangle
        t = (uv * wu - uu * wv) / D;
        if (t < -EPSILON || (s + t-1.0) > EPSILON) continue; //sol is outside triangle

        distances[0] = distance_point_line_3d(sol, tri0, tri1);
        distances[1] = distance_point_line_3d(sol, tri1, tri2);
        distances[2] = distance_point_line_3d(sol, tri2, tri0);
        if(distances[0]>EPSILON2 && distances[1]>EPSILON2 && distances[2]>EPSILON2){
          return 1;
        }
      }

      //test whether ray intersects one of the triangle's edges
      if(intersection_line_line_with_intersection_3d(tri0, tri1, lin0, lin1, sol)){
        return 4;
      }else if(intersection_line_line_with_intersection_3d(tri1, tri2, lin0, lin1, sol)){
        return 2;
      }else if(intersection_line_line_with_intersection_3d(tri2, tri0, lin0, lin1, sol)){
        return 3;
      }
    }
    return 0; //ray disjoint from plane
  }

  //get intersection point of ray with triangle plane
  r = a / b;
  if(r < -EPSILON || r-1.0 > EPSILON){
    return 0; //ray goes away from triangle => no intersection
  }
  for(i=0; i<3; i++) sol[i] = lin0[i] + r * dir[i];

  //is sol inside triangle?
  uu = u[0] * u[0] + u[1] * u[1] + u[2] * u[2];
  uv = u[0] * v[0] + u[1] * v[1] + u[2] * v[2];
  vv = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
  for(i=0; i<3; i++) w[i] = sol[i] - tri0[i];

  wu = w[0] * u[0] + w[1] * u[1] + w[2] * u[2];
  wv = w[0] * v[0] + w[1] * v[1] + w[2] * v[2];
  D = uv * uv - uu * vv;

  //get and test parametric coords
  s = (uv * wv - vv * wu) / D;
  if (s < -EPSILON || s-1.0 > EPSILON){
    return 0; //sol is outside triangle
  }
  t = (uv * wu - uu * wv) / D;
  if (t < -EPSILON || (s + t-1.0) > EPSILON){
    return 0; //sol is outside triangle
  }

  //calculate distance of sol with each edge of the triangle
  distances[0] = distance_point_line_3d(sol, tri0, tri1);
  if(distances[0]<EPSILON2){
    return 4;
  }
  distances[1] = distance_point_line_3d(sol, tri1, tri2);
  if(distances[1]<EPSILON2){
    return 2;
  }
  distances[2] = distance_point_line_3d(sol, tri2, tri0);
  if(distances[2]<EPSILON2){
    return 3;
  }
  return 1; //sol is inside triangle
}

//----------------------------------------< intersection_line_tetrahedron_3d >
bool intersection_line_tetrahedron_3d(double lin0[], double lin1[], double tet0[], double tet1[],
                                                                     double tet2[], double tet3[]){
  double sol[3];
  //test intersection with faces
  if(intersection_line_triangle_3d(tet0, tet1, tet2, lin0, lin1, sol)){
    return true;
  }
  if(intersection_line_triangle_3d(tet0, tet1, tet3, lin0, lin1, sol)){
    return true;
  }
  if(intersection_line_triangle_3d(tet0, tet2, tet3, lin0, lin1, sol)){
    return true;
  }
  if(intersection_line_triangle_3d(tet1, tet2, tet3, lin0, lin1, sol)){
    return true;
  }
  //test intersection with edges
  if(intersection_line_line_3d(tet0, tet1, lin0, lin1)){
    return true;
  }
  if(intersection_line_line_3d(tet0, tet2, lin0, lin1)){
    return true;
  }
  if(intersection_line_line_3d(tet0, tet3, lin0, lin1)){
    return true;
  }
  if(intersection_line_line_3d(tet1, tet2, lin0, lin1)){
    return true;
  }
  if(intersection_line_line_3d(tet1, tet3, lin0, lin1)){
    return true;
  }
  if(intersection_line_line_3d(tet2, tet3, lin0, lin1)){
    return true;
  }
  //test intersection with volume
  if(point_in_tetrahedron_3d(lin0, tet0, tet1, tet2, tet3)
                 || point_in_tetrahedron_3d(lin1, tet0, tet1, tet2, tet3)){
    return true;
  }
  return false;
}

//----------------------------------------< intersection_triangle_tetrahedron_3d >
bool intersection_triangle_tetrahedron_3d(double tri0[], double tri1[], double tri2[],
                                          double tet0[], double tet1[], double tet2[], double tet3[]){
  //test intersection with volume
  if(point_in_tetrahedron_3d(tri0, tet0, tet1, tet2, tet3)
                 || point_in_tetrahedron_3d(tri1, tet0, tet1, tet2, tet3)
                 || point_in_tetrahedron_3d(tri2, tet0, tet1, tet2, tet3)){
    return true;
  }
  //test intersection of triangle-edges with tetrahedron
  if(intersection_line_tetrahedron_3d(tri0, tri1, tet0, tet1, tet2, tet3)){
    return true;
  }
  if(intersection_line_tetrahedron_3d(tri1, tri2, tet0, tet1, tet2, tet3)){
    return true;
  }
  if(intersection_line_tetrahedron_3d(tri2, tri0, tet0, tet1, tet2, tet3)){
    return true;
  }
  //test intersection of tetrahedron-edges with triangle
  double sol[3];
  if(intersection_line_triangle_3d(tri0, tri1, tri2, tet0, tet1, sol)){
    return true;
  }
  if(intersection_line_triangle_3d(tri0, tri1, tri2, tet0, tet2, sol)){
    return true;
  }
  if(intersection_line_triangle_3d(tri0, tri1, tri2, tet0, tet3, sol)){
    return true;
  }
  if(intersection_line_triangle_3d(tri0, tri1, tri2, tet1, tet2, sol)){
    return true;
  }
  if(intersection_line_triangle_3d(tri0, tri1, tri2, tet1, tet3, sol)){
    return true;
  }
  if(intersection_line_triangle_3d(tri0, tri1, tri2, tet2, tet3, sol)){
    return true;
  }
  return false;
}

//----------------------------------------< intersection_box_box_3d >
bool intersection_box_box_3d(double min_corner0[], double max_corner0[],
			     double min_corner1[], double max_corner1[]){
  double p[3];
  
  if (edge_length_3d(min_corner0,max_corner0) < edge_length_3d(min_corner1,max_corner1)) {
  // box0 is smaller than box1
    if (point_in_box_3d(min_corner0, min_corner1, max_corner1))
      return true;
    if (point_in_box_3d(max_corner0, min_corner1, max_corner1))
      return true;
    
    p[0] = min_corner0[0]; p[1] = min_corner0[1]; p[2] = max_corner0[2];
    if (point_in_box_3d(p, min_corner1, max_corner1))
      return true;
    p[0] = min_corner0[0]; p[1] = max_corner0[1]; p[2] = min_corner0[2];
    if (point_in_box_3d(p, min_corner1, max_corner1))
      return true;
    p[0] = min_corner0[0]; p[1] = max_corner0[1]; p[2] = max_corner0[2];
    if (point_in_box_3d(p, min_corner1, max_corner1))
      return true;
    
    p[0] = max_corner0[0]; p[1] = max_corner0[1]; p[2] = min_corner0[2];
    if (point_in_box_3d(p, min_corner1, max_corner1))
      return true;
    p[0] = max_corner0[0]; p[1] = min_corner0[1]; p[2] = max_corner0[2];
    if (point_in_box_3d(p, min_corner1, max_corner1))
      return true;
    p[0] = max_corner0[0]; p[1] = min_corner0[1]; p[2] = min_corner0[2];
    if (point_in_box_3d(p, min_corner1, max_corner1))
      return true;
  
  } else {
  // box1 is smaller than box0
    if (point_in_box_3d(min_corner1, min_corner0, max_corner0))
      return true;
    if (point_in_box_3d(max_corner1, min_corner0, max_corner0))
      return true;
    
    p[0] = min_corner1[0]; p[1] = min_corner1[1]; p[2] = max_corner1[2];
    if (point_in_box_3d(p, min_corner0, max_corner0))
      return true;
    p[0] = min_corner1[0]; p[1] = max_corner1[1]; p[2] = min_corner1[2];
    if (point_in_box_3d(p, min_corner0, max_corner0))
      return true;
    p[0] = min_corner1[0]; p[1] = max_corner1[1]; p[2] = max_corner1[2];
    if (point_in_box_3d(p, min_corner0, max_corner0))
      return true;
    
    p[0] = max_corner1[0]; p[1] = max_corner1[1]; p[2] = min_corner1[2];
    if (point_in_box_3d(p, min_corner0, max_corner0))
      return true;
    p[0] = max_corner1[0]; p[1] = min_corner1[1]; p[2] = max_corner1[2];
    if (point_in_box_3d(p, min_corner0, max_corner0))
      return true;
    p[0] = max_corner1[0]; p[1] = min_corner1[1]; p[2] = min_corner1[2];
    if (point_in_box_3d(p, min_corner0, max_corner0))
      return true;
  }

  return false;
}

//----------------------------------------< intersection_triangle_box_3d >
bool intersection_triangle_box_3d(double tri0[], double tri1[], double tri2[],
				  double min_corner[], double max_corner[]){
  // vertices of triangle covered by box
  if (point_in_box_generous_3d(tri0,min_corner,max_corner))
    return true;
  if (point_in_box_generous_3d(tri1,min_corner,max_corner))
    return true;
  if (point_in_box_generous_3d(tri2,min_corner,max_corner))
    return true;
  
  // bounding box of triangle does not intersect the box
  double p0[3];
  double p1[3];
  boundingbox_triangle_3d(tri0,tri1,tri2,p0,p1);
  if (!intersection_box_box_3d(min_corner,max_corner,p0,p1))
    return false;
  
  double p2[3];
    
  // edges of box intersect triangle
  p0[0] = min_corner[0]; p0[1] = min_corner[1]; p0[2] = max_corner[2];
  if (intersection_line_triangle_3d(tri0,tri1,tri2, min_corner, p0, p2))
    return true;
  
  p1[0] = max_corner[0]; p1[1] = min_corner[1]; p1[2] = max_corner[2];
  if (intersection_line_triangle_3d(tri0,tri1,tri2, max_corner, p1, p2))
    return true;
  if (intersection_line_triangle_3d(tri0,tri1,tri2, p0, p1, p2))
    return true;
  
  p1[0] = min_corner[0]; p1[1] = max_corner[1]; p1[2] = max_corner[2];
  if (intersection_line_triangle_3d(tri0,tri1,tri2, max_corner, p1, p2))
    return true;
  if (intersection_line_triangle_3d(tri0,tri1,tri2, p0, p1, p2))
    return true;
  
  p0[0] = min_corner[0]; p0[1] = max_corner[1]; p0[2] = min_corner[2];
  if (intersection_line_triangle_3d(tri0,tri1,tri2, min_corner, p0, p2))
    return true;
  if (intersection_line_triangle_3d(tri0,tri1,tri2, p0, p1, p2))
    return true;
  
  p1[0] = max_corner[0]; p1[1] = max_corner[1]; p1[2] = min_corner[2];
  if (intersection_line_triangle_3d(tri0,tri1,tri2, max_corner, p1, p2))
    return true;
  if (intersection_line_triangle_3d(tri0,tri1,tri2, p0, p1, p2))
    return true;
  
  p0[0] = max_corner[0]; p0[1] = min_corner[1]; p0[2] = min_corner[2];
  if (intersection_line_triangle_3d(tri0,tri1,tri2, min_corner, p0, p2))
    return true;
  if (intersection_line_triangle_3d(tri0,tri1,tri2, p0, p1, p2))
    return true;
  
  p1[0] = max_corner[0]; p1[1] = min_corner[1]; p1[2] = max_corner[2];
  if (intersection_line_triangle_3d(tri0,tri1,tri2, p0, p1, p2))
    return true;
  
  return false;
}

//----------------------------------------< distance_point_line_3d >
//calculates the distance between a given point and the line segment between 'lin0' and 'lin1'.
double distance_point_line_3d(double point[], double lin0[], double lin1[]){
  double gx, gy, gz, fx, fy, fz, dx, dy, dz, t;

  //the (complete) line through 'lin0' and 'lin1' is given by 'l = lin0 + t*g' with g as follows:
  gx = lin1[0] - lin0[0];
  gy = lin1[1] - lin0[1];
  gz = lin1[2] - lin0[2];

  //the normal-vector from 'point' to l is 'v = lin0 + t*g - point' for some t which is determined
  // by 'v*g = 0'
  t = (-gx * (lin0[0]-point[0]) - gy * (lin0[1]-point[1]) - gz * (lin0[2]-point[2]))
                                                            / (gx*gx + gy*gy + gz*gz);

  //if the orthogonal projection of 'point' onto l is not on the given line segment then one of the
  //vertices 'lin0' or 'lin1' is the nearest point to 'point'
  if(t<0){
    //'lin0' is nearest
    fx = lin0[0];
    fy = lin0[1];
    fz = lin0[2];
  }
  else if(t>1){
    //'lin1' is nearest
    fx = lin1[0];
    fy = lin1[1];
    fz = lin1[2];
  }
  else{
    //orthogonal projection is nearest
    fx = lin0[0] + t * gx;
    fy = lin0[1] + t * gy;
    fz = lin0[2] + t * gz;
  }

  //calculate distance
  dx = point[0] - fx;
  dy = point[1] - fy;
  dz = point[2] - fz;
  return sqrt(dx*dx + dy*dy + dz*dz);
}


//----------------------------------------< distance_point_box_2d >
double distance_point_box_3d(double point[], double min_corner[], double max_corner[]){
  if (point_in_box_3d(point,min_corner,max_corner))
    return 0.0;
  
  double p[3];
  double dist = edge_length_3d(point, min_corner);  
  dist = std::min(dist, edge_length_3d(point, min_corner)); 
  
  p[0] = min_corner[0]; p[1] = min_corner[1]; p[2] = max_corner[2];
  dist = std::min(dist, edge_length_3d(point, p));  
  p[0] = min_corner[0]; p[1] = max_corner[1]; p[2] = min_corner[2];
  dist = std::min(dist, edge_length_3d(point, p));
  p[0] = min_corner[0]; p[1] = max_corner[1]; p[2] = max_corner[2];
  dist = std::min(dist, edge_length_3d(point, p));
  
  p[0] = max_corner[0]; p[1] = max_corner[1]; p[2] = min_corner[2];
  dist = std::min(dist, edge_length_3d(point, p));  
  p[0] = max_corner[0]; p[1] = min_corner[1]; p[2] = max_corner[2];
  dist = std::min(dist, edge_length_3d(point, p));
  p[0] = max_corner[0]; p[1] = min_corner[1]; p[2] = min_corner[2];
  dist = std::min(dist, edge_length_3d(point, p));
  
  return dist;
}

//----------------------------------------< distance_point_triangle_3d >
//calculates the distance between a given 'point' and the triangle given by 'tri0', 'tri1', 'tri2'.
double distance_point_triangle_3d(double point[], double tri0[], double tri1[], double tri2[]){
  double gx, gy, gz, hx, hy, hz, fx, fy, fz, u, v;
  double c1, c2, b1, b2, a1, distance, tmp;

  //triangle lies in plane 'p = tri0 + u*g + v*h', where g and h are defined as follows:
  gx = tri1[0] - tri0[0];
  gy = tri1[1] - tri0[1];
  gz = tri1[2] - tri0[2];
  hx = tri2[0] - tri0[0];
  hy = tri2[1] - tri0[1];
  hz = tri2[2] - tri0[2];

  //a vector from 'point' to p can be described by 'l = tri0 + u*g + v*h - point' and since we search the
  //distance we need such a vector that is orthogonal to p. Thus 'l*g = 0' and 'l*h = 0'. We get a system
  //of two equations of the form:
  //a1*u + b1*v = c1
  //b1*u + b2*v = c2
  //with a1, b1, b2, c1, c2 as follows:
  b1 = gx*hx + gy*hy + gz*hz;
  c2 = -hx*(tri0[0]-point[0]) - hy*(tri0[1]-point[1]) - hz*(tri0[2]-point[2]);
  b2 = hx*hx + hy*hy + hz*hz;
  c1 = -gx*(tri0[0]-point[0]) - gy*(tri0[1]-point[1]) - gz*(tri0[2]-point[2]);
  a1 = gx*gx + gy*gy + gz*gz;

  //solve the system:
  if(abs(b1) < EPSILON){
    v = c2/b2;
    u = c1/a1;
  }
  else{
    v = (c1*b1 - c2*a1) / (b1*b1 - b2*a1);
    u = (c1 - v*b1) / a1;
  }

  //now test whether the orthogonal projection of 'point' onto p lies inside the triangle
  if(u>=0 && v>=0 && u+v<=1){
    //yes, inside. The length of the normal-vector is the desired distance
    fx = point[0] - tri0[0] - u*gx - v*hx;
    fy = point[1] - tri0[1] - u*gy - v*hy;
    fz = point[2] - tri0[2] - u*gz - v*hz;
    distance = sqrt(fx*fx + fy*fy + fz*fz);
  }
  else{
    //no, not inside. The desired distance is the distance to one of the triangle-edges
    distance = 9999999.0;
    tmp = distance_point_line_3d(point, tri0, tri1);
    if(tmp<distance) distance = tmp;
    tmp = distance_point_line_3d(point, tri1, tri2);
    if(tmp<distance) distance = tmp;
    tmp = distance_point_line_3d(point, tri2, tri0);
    if(tmp<distance) distance = tmp;
  }
  return distance;
}


double volume_tetrahedron(double tet0[], double tet1[], double tet2[], double tet3[]){
  double det = solve_determinant3(
    tet1[0]-tet0[0], tet2[0]-tet0[0], tet3[0]-tet0[0],
    tet1[1]-tet0[1], tet2[1]-tet0[1], tet3[1]-tet0[1],
    tet1[2]-tet0[2], tet2[2]-tet0[2], tet3[2]-tet0[2]);
    
  return abs(det)/6.0;
}

}
