/** \file VectorOperations.h */

#ifndef VECTOR_OPERATIONS_H
#define VECTOR_OPERATIONS_H

#include "AMDiS.h"

#if HAVE_PARALLEL_DOMAIN_AMDIS
#include "parallel/StdMpi.h"
#endif

#include <unistd.h>
#include <ios>
#include <iostream>
#include <fstream>
#include <string>
#include <boost/numeric/mtl/mtl.hpp>
#include <boost/numeric/itl/itl.hpp>

using namespace AMDiS; 

template<typename T>
const WorldVector<T>& operator*=(WorldVector<T>& v1, const WorldVector<T>& v2)
{
  T* v1It, *v2It;
  for (v1It = v1.begin(), v2It = v2.begin();
	  v1It != v1.end(); 
	  v1It++, v2It++)
  *v1It *= *v2It;
  
  return v1;
};

// x ~= y
template<typename T>
class compareTol : public binary_function<T, T, bool> {
public:
	compareTol(double tol_) : tol(tol_) {}
	bool operator()(const T &x, const T &y) const { return sqrt((x-y)*(x-y))<tol; }
	double tol;
};
template<typename T1, typename T2, int component>
struct comparePair : public binary_function<std::pair<T1, T2>, std::pair<T1, T2>, bool> 
{
  bool operator()(const std::pair<T1, T2> &x, const std::pair<T1, T2> &y) const 
  { 
    return (component==0?(x.first<y.first):(x.second<y.second));
  }
};

template<typename T1, typename T2>
struct compare0 : public binary_function<std::pair<T1, T2>, std::pair<T1, T2>, bool> 
{
  bool operator()(const std::pair<T1, T2> &x, const std::pair<T1, T2> &y) const 
  { 
    return x.first<y.first;
  }
};
template<typename T1, typename T2>
struct compare1 : public binary_function<std::pair<T1, T2>, std::pair<T1, T2>, bool> 
{
  bool operator()(const std::pair<T1, T2> &x, const std::pair<T1, T2> &y) const 
  { 
    return x.second<y.second;
  }
};

class SortPair : public binary_function<std::pair<WorldVector<double>, double>, std::pair<WorldVector<double>, double>, bool> {
public:
	SortPair(int component_=1,int idx_=0) : component(component_),idx(idx_) {}
	bool operator()(const std::pair<WorldVector<double>, double> &x, 
		const std::pair<WorldVector<double>, double> &y) const { 
		return (component==1?(x.first)[idx]<(y.first)[idx]:x.second<y.second); }
	int component,idx;
};

namespace vector_operations {

  // num_rows for vector types
  template<typename T>
  size_t num_rows(WorldVector<T> &v) {
    return static_cast<size_t>(v.getSize());
  };
  template<typename T>
  size_t num_rows(const WorldVector<T> &v) {
    return static_cast<size_t>(v.getSize());
  };
  template<typename T>
  size_t num_rows(std::vector<T> &v) {
    return static_cast<size_t>(v.size());
  };
  template<typename T>
  size_t num_rows(const std::vector<T> &v) {
    return static_cast<size_t>(v.size());
  };
  template<typename T>
  size_t num_rows(T &v) {
    return 1;
  };
  template<typename T>
  size_t num_rows(const T &v) {
    return 1;
  };

  // num_rows/num_cols for matrix types
  template<typename T>
  size_t num_rows(WorldMatrix<T> &m) {
    return static_cast<size_t>(m.getNumRows());
  };
  template<typename T>
  size_t num_rows(const WorldMatrix<T> &m) {
    return static_cast<size_t>(m.getNumRows());
  };
  template<typename T>
  size_t num_cols(WorldMatrix<T> &m) {
    return static_cast<size_t>(m.getNumCols());
  };
  template<typename T>
  size_t num_cols(const WorldMatrix<T> &m) {
    return static_cast<size_t>(m.getNumCols());
  };
  
  // common interface to resize vectors
  template<typename T>
  void resize(WorldVector<T> &v, size_t dim) {
    TEST_EXIT(dim==v.getSize())("WorldVectors can not be resized to the given dimension!\n");
  };
  template<typename T>
  void resize(std::vector<T> &v, size_t dim) {
    v.resize(dim);
  };
  template<typename T>
  void resize(mtl::dense_vector<T> &v, size_t dim) {
    v.change_dim(dim);
  };
  
  
  
  // B(phi)=phi^2*(1-phi)^2
  template<typename T>
  T norm(mtl::dense_vector<T> &b) { return two_norm(b); };
  
  template<typename T>
  T norm(std::vector<T> &b) 
  {
    T value = 0.0;
    for (size_t i = 0; i < b.size(); ++i)
      value += sqr(b[i]);
    return sqrt(value);
  };
  
  template<typename Matrix>
  double matrix_norm(Matrix &M)
  {
    typename Matrix::value_type asum;
    asum = 0.0;
    for (size_t i = 0; i < num_rows(M); ++i) {
      for (size_t j = 0; j < num_cols(M); ++j)
	asum += abs(M[i][j]);
    }
    return static_cast<double>(asum/(num_rows(M)*num_cols(M)));
  };
  
  // |x|_eps
  inline double epsNorm(const WorldVector<double> &x, const double &eps) { return std::max(AMDiS::norm(x),eps); };
  
  // 1/|x|_eps
  inline double epsNormInv(const WorldVector<double> &x, const double &eps) { return 1.0/epsNorm(x,eps); };
  
  // 1/[|x|_eps]^3
  inline double epsNormInv3(const WorldVector<double> &x, const double &eps)
  {
    double result = epsNorm(x,eps);
    result = 1.0/(result*sqr(result));  
    return result;
  };
  
  // y:=x/|x|_eps
  inline void epsNormalize(const WorldVector<double> &x, const double &eps, WorldVector<double> &y)
  {
    double nrmInv = epsNormInv(x,eps);
    for(int k=0; k<x.getSize(); k++) {
      y[k] = x[k] * nrmInv;
    }
  };

  template<typename Matrix>
  double trace(Matrix &M)
  {
    TEST_EXIT(num_rows(M)==num_cols(M))("Matrix dimensions must be equal!\n");
    typename Matrix::value_type sum;
    sum = 0.0;
    for (size_t i = 0; i < num_rows(M); ++i) {
      sum += M[i][i];
    }
    return static_cast<double>(sum);
  };
  
  template<typename T>
  Vector<T> mult(const Vector<T>& v1, const Vector<T>& v2)
  {
    TEST_EXIT(v1.getSize()==v2.getSize())("Vectors must have the same size!\n");
    Vector<T> result(v1.getSize());
    Vector<T> *v1It, *v2It, *resultIt;
    for (v1It = v1.begin(), v2It = v2.begin(), resultIt = result.begin(); v1It != v1.end(); ++v1It, ++v2It, ++resultIt)
	*resultIt = *v1It * *v2It;
    return result;
  };
  
  inline WorldVector<double> mult(const mtl::dense2D<double> &A, const WorldVector<double> &x)
  {
    TEST_EXIT(num_rows(A)==x.getSize())("Falsche Dimensionen multipliziert!");
    WorldVector<double> result;
    for (unsigned i = 0; i < x.getSize(); ++i) {
      result[i] = 0.0;
      for (unsigned j = 0; j < x.getSize(); ++j)
	result[i] += A[i][j]*x[j];
    }
    return result;
  };
  
  template<typename ResultVector, typename Matrix, typename Vector1, typename Vector2>
  ResultVector Axpy(const Matrix &A, const Vector1 &x, const Vector2 &y)
  {
    size_t n = num_rows(x);
    TEST_EXIT(num_rows(A)==n && num_cols(A)==n && num_rows(y)==n)("Falsche Dimensionen multipliziert!");
    ResultVector result;
    resize(result, n);
    for (unsigned i = 0; i < n; ++i) {
      result[i] = y[i];
      for (unsigned j = 0; j < n; ++j)
	result[i] += A[i][j]*x[j];
    }
    return result;
  };

  inline void unitVector(WorldVector<double> &b, int &idx) {
    TEST_EXIT(idx>=0 && idx<b.getSize())("Unit-Vector component does not exist!\n");
    b.set(0.0);
    b[idx]=1.0;
  };
  
  inline WorldVector<double> *unitVector(int &idx) {
    WorldVector<double> *result = new WorldVector<double>;
    TEST_EXIT(idx>=0 && idx<result->size())("Unit-Vector component does not exist!\n");
    (*result)[idx]=1.0;
    return result;
  };

  template<typename T>
  void copy(std::vector<T> &v, WorldVector<T> &w)
  {
    for(int i=0; i<std::min((int)v.size(),(int)w.getSize()); i++)
	    w[i]=v[i];
  };
  
  template<typename T>
  void copy(mtl::dense_vector<T> &v, WorldVector<T> &w)
  {
    for(int i=0; i<std::min((int)num_rows(v),(int)w.getSize()); i++)
	    w[i]=v[i];
  };
  
  template<typename T>
  void copy(WorldVector<T> &w, mtl::dense_vector<T> &v)
  {
    for(int i=0; i<std::min((int)num_rows(v),(int)w.getSize()); i++)
	    v[i]=w[i];
  };
  
  template<typename T>
  void merge(std::vector<T> &v1, std::vector<T> &v2, std::vector<T> &v3)
  {
    v3.resize(v1.size()+v2.size());
    for(int i=0;i<v1.size();i++)
	    v3[i]=v1[i];
    for(int j=0;j<v2.size();j++)
	    v3[j+v1.size()]=v2[j];
  };
  
  template<typename T>
  void getMin(mtl::dense_vector<T> &v, T &minVal, unsigned &minIdx)
  {
    if(num_rows(v)==0) {
      minVal = numeric_limits<T>::infinity();
    } else {
      minVal=v[0]; minIdx=0;
      for(unsigned i=1; i<num_rows(v); i++) {
	if(v[i]<minVal) { minVal=v[i]; minIdx=i; }
      }
    }
  };
  
  template<typename T>
  void getMin(std::vector<T> &v, T &minVal, size_t &minIdx)
  {
    if (v.size() == 0) {
      minVal = numeric_limits<T>::infinity();
    } else {
      minVal = v[0];
      minIdx = 0;
      for(size_t i = 1; i < v.size(); i++) {
	if (v[i] < minVal) { 
	  minVal=v[i];
	  minIdx=i;
	}
      }
    }
  };
  
  template<typename T>
  void getMax(mtl::dense_vector<T> &v, T &maxVal, unsigned &maxIdx)
  {
    if(num_rows(v)==0) {
      maxVal = -numeric_limits<T>::infinity();
    } else {
      maxVal=v[0]; maxIdx=0;
      for(unsigned i=1; i<num_rows(v); i++) {
	if(v[i]>maxVal) { maxVal=v[i]; maxIdx=i; }
      }
    }
  };

  template<typename T>
  T max(const std::vector<T> &vec)
  {
    double maxVal;
    if(vec.size()==0) {
      maxVal = -numeric_limits<T>::infinity();
    } else {
      maxVal=vec[0];
      for(size_t i=1; i<vec.size(); i++) {
	if(vec[i]>maxVal) { maxVal=vec[i]; }
      }
    }
    return maxVal;
  };

  template<typename T>
  T min(const std::vector<T> &vec)
  {
    double minVal;
    if(vec.size()==0) {
      minVal = numeric_limits<T>::infinity();
    } else {
      minVal=vec[0];
      for(size_t i=1; i<vec.size(); i++) {
	if(vec[i]<minVal) { minVal=vec[i]; }
      }
    }
    return minVal;
  };
  
  template<typename T>
  T absmax(std::vector<T> &vec)
  {
    double maxVal;
    if(vec.size()==0) {
      maxVal = -numeric_limits<T>::infinity();
    } else {
      maxVal=abs(vec[0]);
      for(size_t i=1; i<vec.size(); i++) {
	if(abs(vec[i])>maxVal) { maxVal=abs(vec[i]); }
      }
    }
    return maxVal;
  };
  
  template<typename T>
  T sum(std::vector<T> &vec)
  {
    unsigned N=vec.size();
    T value = 0;
    if(N>0) {
      value = vec[0];
      for(unsigned i=1; i<N; ++i) {
	value += vec[i];
      }
    }
    return value;
  };
  
  template<typename T>
  void getMean(std::vector<T> &vec, T &meanValue)
  {
    unsigned N=vec.size();
    meanValue = sum(vec)/double(N);
  };
  
//   template<typename Vector1, typename Vector2>
//   struct VectorPair {
//     typedef typename Vector1::size_type 	size_type;
//     typedef typename Vector1::value_type 	T1;
//     typedef typename Vector2::value_type 	T2;
//     typedef typename Vector1::iterator 		iterator1;
//     typedef typename Vector2::iterator 		iterator2;
//     typedef typename std::pair<T1, T2>& 	reference;
//     typedef typename const std::pair<T1, T2>& 	const_reference;
//     
//     typedef struct PairIterator : public iterator<random_access_iterator_tag> {
//       PairIterator(iterator1& iter1_, iterator2& iter2_) : iter1(iter1_), iter2(iter2_) {}
//       PairIterator(const PairIterator& piter) : iter1(piter.iter1), iter2(piter.iter2) {}
//       PairIterator& operator++() {++iter1; ++iter2; return *this;}
//       PairIterator& operator--() {--iter1; --iter2; return *this;}
//       PairIterator& operator+(int n) {PairIterator tmp(*this); tmp.iter1+=n; tmp.iter2+=n; return tmp;}
//       PairIterator& operator-(int n) {PairIterator tmp(*this); tmp.iter1-=n; tmp.iter2-=n; return tmp;}
//       PairIterator& operator+=(int n) {iter1+=n; iter2+=n; return *this;}
//       PairIterator& operator-=(int n) {iter1-=n; iter2-=n; return *this;}
// //       PairIterator operator++(int) {PairIterator tmp(*this); operator++(); return tmp;} //?1
//       bool operator==(const PairIterator& rhs) {return iter1 == rhs.iter1 && iter2 == rhs.iter2;}
//       bool operator<(const PairIterator& rhs) {return iter1 < rhs.iter1 && iter2 < rhs.iter2;}
//       bool operator>(const PairIterator& rhs) {return iter1 > rhs.iter1 && iter2 > rhs.iter2;}
//       bool operator<=(const PairIterator& rhs) {return iter1 <= rhs.iter1 && iter2 <= rhs.iter2;}
//       bool operator>=(const PairIterator& rhs) {return iter1 >= rhs.iter1 && iter2 >= rhs.iter2;}
//       bool operator!=(const PairIterator& rhs) {return iter1 != rhs.iter1 || iter2 != rhs.iter2;}
//       std::pair<T1&, T2&>& operator*() {return std::make_pair(*iter1, *iter2);}
//       std::pair<T1&, T2&>& operator[](int n) {return std::make_pair(*(iter1+n), *(iter2+n));}
// 
//     protected:
//       iterator1 iter1;
//       iterator2 iter2;
//     } iterator;
//   };
  
  template<typename T1, typename T2, typename Compare>
  void sort(std::vector<T1> &vec1, std::vector<T2> &vec2, Compare &comp)
  {
    std::vector<pair<T1,T2> > order(vec1.size());
    
    // create vector of pairs
    for (int i=0; i<vec1.size(); ++i)
      order[i] = make_pair(vec1[i], vec2[i]);
    
    // sort vector of pairs
    sort(order.begin(), order.end(), comp);
    
    // copy sorted pairs back to vectors
    for (int i=0; i<vec1.size(); ++i) {
      vec1[i] = order[i].first;
      vec2[i] = order[i].second;
    }
  };
  
  template<typename T1, typename T2, typename Compare>
  void sort(std::vector<std::pair<T1,T2> > &vec, Compare &comp)
  {
    std::sort(vec.begin(), vec.end(), comp);
  };
  
  inline void unique(std::vector<WorldVector<double> > &vec, double tol, std::vector<unsigned> &ind)
  {
    compareTol<WorldVector<double> > comp(tol);
    unsigned newVec=0;
    for(unsigned i=0; i<vec.size();++i) {
      bool inNew=false;
      for(unsigned j=0;j<newVec;++j) {
	inNew=inNew||comp(vec[i],vec[j]);
	if(inNew) break;
      }
      if(!inNew) {vec[newVec]=vec[i]; newVec++; ind.push_back(i);}
    }
    vec.erase(vec.begin()+newVec,vec.end());
  };
  
  template<typename T1, typename T2>
  void unique(std::vector<std::pair<T1,T2> > &vec, double tol, int uniqueBy=1)
  {
//     comparePair comp = comparePair(tol,uniqueBy);
//     unsigned newVec=0;
//     for(unsigned i=0; i<vec.size();++i) {
//       bool inNew=false;
//       for(unsigned j=0;j<newVec;++j) {
// 	inNew=inNew||comp(vec[i],vec[j]);
// 	if(inNew) break;
//       }
//       if(!inNew) {vec[newVec]=vec[i]; newVec++;}
//     }
//     vec.erase(vec.begin()+newVec,vec.end());
  };
  
  inline void filter(std::vector<WorldVector<double> > &x, std::vector<bool> &filter)
  {
    unsigned j=0;
    for(unsigned i=0; i<x.size(); ++i) {
      if(filter[i]) {
	x[j]=x[i];
	j++;
      }
    }
    x.erase(x.begin()+j,x.end());
  };
  
  inline void filter(std::vector<WorldVector<double> > &x, std::vector<int> &filter)
  {
    unsigned j=0;
    for(unsigned i=0; i<filter.size(); ++i) {
      x[j]=x[filter[i]];
      j++;
    }
    x.erase(x.begin()+j,x.end());
  };
  template<typename T, typename T1, typename T2>
  T dot(std::vector<T1> &vec1, std::vector<T2> &vec2)
  {
    unsigned N=vec1.size();
    T value = 0;
    if(N>0) {
      value = static_cast<T>(vec1[0])*static_cast<T>(vec2[0]);
      for(unsigned i=1; i<N; ++i) {
	value += static_cast<T>(vec1[i])*static_cast<T>(vec2[i]);
      }
    }
    return value;
  };
  template<typename T, typename T1, typename T2>
  T dot(mtl::dense_vector<T1> &vec1, WorldVector<T2> &vec2)
  {
      unsigned N=vec2.getSize();
      T value = 0;
      if(N>0) {
	  value = static_cast<T>(vec1[0])*static_cast<T>(vec2[0]);
	  for(unsigned i=1; i<N; ++i) {
	      value += static_cast<T>(vec1[i])*static_cast<T>(vec2[i]);
	  }
      }
      return value;
  };
  
  template<class T>
  static void fillMatrix(WorldMatrix<T> &m, mtl::dense2D<T> &m_in)
  {
    m.resize(num_rows(m_in),num_cols(m_in));
    for (size_t i = 0; i < num_rows(m_in); ++i) {
      for (size_t j = 0; j < num_cols(m_in); ++j) {
	m[i][j] = m_in(i,j);
      }
    }
  };
}

#endif // VECTOR_OPERATIONS_H
