/** \file CahnHilliard.h */

#ifndef CAHN_HILLIARD_H
#define CAHN_HILLIARD_H

#include "AMDiS.h"
#include "BaseProblem.h"
#include "ExtendedProblemStat.h"
#include "HL_SignedDistTraverse.h"

using namespace AMDiS;

class CahnHilliard : public BaseProblem<ExtendedProblemStat>
{
public: // definition of types

  typedef BaseProblem<ExtendedProblemStat> super;

public: // public methods

  CahnHilliard(const std::string &name_);
  ~CahnHilliard();

  virtual void initData();
  virtual void solveInitialProblem(AdaptInfo *adaptInfo);

  double getEpsilon() { return eps; }
  int getDoubleWellType() { return doubleWell; }

protected: // protected methods

  virtual void fillOperators();
  virtual void fillBoundaryConditions() {}

protected: // protected variables

  HL_SignedDistTraverse *reinit;

  bool useMobility;

  unsigned dim;

  int doubleWell;

  double gamma;
  double eps;
  double minusEps;
  double epsInv;
  double minusEpsInv;
  double epsSqr;
  double minusEpsSqr;
};


/** \brief
 * Abstract function for Cahn-Hilliard mobility
 */
class MobilityCH0 : public AbstractFunction<double,double>
{
  public:
    MobilityCH0(double gamma_=1.0) : 
      AbstractFunction<double,double>(4), 
      gamma(gamma_),
      delta(1.e-6) { }
    double operator()(const double &ch) const 
    {
      double phase = std::max(0.0, std::min(1.0, ch));
      double mobility = 0.25*sqr(phase)*sqr(phase-1.0);
      return gamma * std::max(mobility, delta);
    }

  protected:
    double gamma;
    double delta;
};

class MobilityCH1 : public AbstractFunction<double,double>
{
  public:
    MobilityCH1(double gamma_=1.0) :
      AbstractFunction<double,double>(4),
      gamma(gamma_),
      delta(1.e-6) { }
    double operator()(const double &ch) const
    {
      double phase = std::max(-1.0, std::min(1.0, ch));
      double mobility = 0.25*sqr(sqr(phase)-1.0);
      return gamma * std::max(mobility, delta);
    }

  protected:
    double gamma;
    double delta;
};

class ScaleFunctor : public AbstractFunction<double,double>
{
  public:
    ScaleFunctor(double factor_=1.0) :
      AbstractFunction<double,double>(1),
      factor(factor_) { }
    double operator()(const double &ch) const
    {
      return factor * ch;
    }

  protected:
    double factor;
};

class Pow2Functor : public AbstractFunction<double,double>
{
  public:
    Pow2Functor(double factor_=1.0) :
      AbstractFunction<double,double>(2),
      factor(factor_) { }
    double operator()(const double &ch) const
    {
      return factor * sqr(ch);
    }

  protected:
    double factor;
};

class Pow3Functor : public AbstractFunction<double,double>
{
  public:
    Pow3Functor(double factor_=1.0) :
      AbstractFunction<double,double>(3),
      factor(factor_) { }
    double operator()(const double &ch) const
    {
      return factor * sqr(ch) * ch;
    }

  protected:
    double factor;
};

#endif // CAHN_HILLIARD_H
