#include "NavierStokesPhase_Chorin.h"

#define UPDATE_VELOCITY_SYSTEM

using namespace std;
using namespace AMDiS;

NavierStokesPhase_Chorin::NavierStokesPhase_Chorin(const std::string &name_, WorldVector<AbstractFunction<double, WorldVector<double> >* > initialVelocityFcts_, DiffuseDomain* geometryProb_) :
  super(name_, initialVelocityFcts_),
  geometryProb(geometryProb_),
  beta(1.0),
  epsPhase2(1.0)
{
  Initfile::get(name + "->beta", beta); // penalization of boundary condition
  double exponent = 2.7;
  Initfile::get(name + "->exponent", exponent); // penalization of boundary condition
  epsPhase2 = beta/pow(geometryProb->getEpsilon(), exponent);

  writer = new MyFileWriter();
  writer->initFile("diff");
};


void NavierStokesPhase_Chorin::solveInitialProblem(AdaptInfo *adaptInfo) 
{ FUNCNAME("NavierStokesPhase_Chorin::solveInitialProblem()");
MSG_DBG("NavierStokesPhase_Chorin::solveInitialProblem()\n");

  NavierStokes_Chorin::solveInitialProblem(adaptInfo);

  if (initialVelocityIsSet) {
  for (unsigned i = 0; i < dim; ++i) {
    transformDOF(predictorProb->getSolution()->getDOFVector(i),
      geometryProb->getPhasefield(), 
      predictorProb->getSolution()->getDOFVector(i), 
      new Mult<double>);
  }
  }
  tau = 0.1;

};


void NavierStokesPhase_Chorin::fillOperators()
{ FUNCNAME("NavierStokesPhase_Chorin::fillOperators()");
  MSG_DBG("NavierStokesPhase_Chorin::fillOperators()\n");

  WorldVector<DOFVector<double>* > vel;
  for (unsigned i = 0; i < dim; i++)
    vel[i]= correctorProb->getSolution()->getDOFVector(i);

  // fill operators for predictorProb
  for (unsigned i = 0; i < dim; ++i) {
    /// < (1/tau)*u'_i , psi >
    Operator *opTime = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(i));
    opTime->addZeroOrderTerm(new Phase_ZOT(geometryProb->getPhasefield()));
    predictorProb->addMatrixOperator(*opTime, i, i, getInvTau());
    /// < (1/tau)*u_i^old , psi >
    opTime->setUhOld(correctorProb->getSolution()->getDOFVector(i));
    predictorProb->addVectorOperator(*opTime, i, getInvTau());
 
    // boundary condition: u=0 on Gamma
    /// < (beta/eps^3)*(1-phi)*u'_i , psi >
    Operator *bound0 = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(i));
    bound0->addTerm(new PhaseInverse_ZOT(
      geometryProb->getPhasefield(),
      epsPhase2));
    predictorProb->addMatrixOperator(*bound0, i, i);
    bound0->setUhOld(boundaryDOF[i]);
    predictorProb->addVectorOperator(*bound0, i);

    if (nonLinTerm == 0) {
      /// < phi*u^old*grad(u_i^old) , psi >
      Operator *opUGradU = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(i));
      opUGradU->addTerm(new WorldVecPhase_FOT(geometryProb->getPhasefield(), vel, -1.0), GRD_PHI);
      opUGradU->setUhOld(correctorProb->getSolution()->getDOFVector(i));
      predictorProb->addVectorOperator(*opUGradU, i);
    } else if (nonLinTerm == 1) {
      /// < phi*u'*grad(u_i^old) , psi >
      for (unsigned j = 0; j < dim; ++j) {
        Operator *opUGradU = new Operator(predictorProb->getFeSpace(i),predictorProb->getFeSpace(i));
        opUGradU->addTerm(new VecAndPartialDerivative_ZOT(
          geometryProb->getPhasefield(), 
          correctorProb->getSolution()->getDOFVector(i), j));
        predictorProb->addMatrixOperator(*opUGradU, i, j);
      }
    } else {
      /// < phi*u^old*grad(u'_i) , psi >
      for (unsigned j = 0; j < dim; ++j) {
        Operator *opUGradU = new Operator(predictorProb->getFeSpace(i),predictorProb->getFeSpace(i));
        opUGradU->addTerm(new Vec2ProductPartial_FOT(
          geometryProb->getPhasefield(), 
          correctorProb->getSolution()->getDOFVector(j),
          j), GRD_PHI);
        predictorProb->addMatrixOperator(*opUGradU, i, i);
      }           
    }

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) > TODO: Add boundary-integral for neumann-boundary
    Operator *opGradP2 = new Operator(predictorProb->getFeSpace(i),pressureProb->getFeSpace(0));
    opGradP2->addFirstOrderTerm(new VecAndPartialDerivative_FOT(geometryProb->getPhasefield(), i, -1.0), GRD_PHI);
    opGradP2->setUhOld(pressureProb->getSolution()->getDOFVector(0));
    predictorProb->addVectorOperator(*opGradP2, i, getPressureStep());
  
      
    /// external force, i.e. gravitational force
    if (norm(force) > DBL_TOL) {
      Operator *opForce = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(i));
      opForce->addZeroOrderTerm(new Phase_ZOT(geometryProb->getPhasefield(),force[i]));
      predictorProb->addVectorOperator(*opForce, i);
    }
  }

  // preassure-calculation: tau*div(phi*grad(p)) = div(phi*u')-grad(phi)*g
  // ---------------------------------------------------------------------
  for(unsigned i = 0; i < dim; ++i) {
    /// < (1/tau)*phi*d_i(u'_i) , psi >
    Operator *opDivU = new Operator(pressureProb->getFeSpace(0),predictorProb->getFeSpace(i));
    opDivU->addTerm(new VecAndPartialDerivative_ZOT(
      geometryProb->getPhasefieldDisturbed(),
      predictorProb->getSolution()->getDOFVector(i),i));
    pressureProb->addVectorOperator(opDivU,0);
    /// < 1/tau*d_i(phi)*u'_i , psi >
    Operator *gradPhiU = new Operator(pressureProb->getFeSpace(0), predictorProb->getFeSpace(i));
    gradPhiU->addTerm(new VecAndPartialDerivative_ZOT(
      predictorProb->getSolution()->getDOFVector(i),
      geometryProb->getPhasefieldDisturbed(),i));
    pressureProb->addVectorOperator(gradPhiU,0);
    /// < 1/tau*d_i(phi)*g , psi >
    Operator *gradPhiG = new Operator(pressureProb->getFeSpace(0), predictorProb->getFeSpace(i));
    gradPhiG->addTerm(new VecAndPartialDerivative_ZOT(
      boundaryDOF[i],
      geometryProb->getPhasefieldDisturbed(),i,-1.0));
    pressureProb->addVectorOperator(gradPhiG,0);
  }
  /// < -grad(p) , grad(psi) >
  Operator *laplaceP = new Operator(pressureProb->getFeSpace(0), pressureProb->getFeSpace(0));
  laplaceP->addSecondOrderTerm(new Phase_SOT(geometryProb->getPhasefieldDisturbed(), -1.0));
  pressureProb->addMatrixOperator(*laplaceP, 0, 0, getTau());

  // corrector step: u = u' - tau*grad(p)
  // ---------------------------------------
  for(unsigned i = 0; i < dim; ++i) {
    /// < u_i , psi >
    Operator *opSimple = new Operator(correctorProb->getFeSpace(i),correctorProb->getFeSpace(i));
    opSimple->addZeroOrderTerm(new Simple_ZOT);
    correctorProb->addMatrixOperator(*opSimple, i, i);
    /// < u'_i , psi >
    opSimple->setUhOld(predictorProb->getSolution()->getDOFVector(i));
    correctorProb->addVectorOperator(*opSimple, i);
    /// < -tau*d_i(p) , psi >
    Operator *opGradQ = new Operator(correctorProb->getFeSpace(i),pressureProb->getFeSpace(0));
    opGradQ->addFirstOrderTerm(new PartialDerivative_FOT(i, -1.0), GRD_PHI);
    opGradQ->setUhOld(pressureProb->getSolution()->getDOFVector(0));
    correctorProb->addVectorOperator(*opGradQ, i, getTau());
  }
};


void NavierStokesPhase_Chorin::addLaplaceTerm(int i)
{
  MSG_DBG("NavierStokesPhase_Chorin::addLaplaceTerm()\n");
    /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
    if (laplaceType == 1) {
      for (unsigned j = 0; j < dim; ++j) {
        Operator *opLaplaceUi1 = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(j));
        opLaplaceUi1->addSecondOrderTerm(new MatrixIJPhase_SOT(geometryProb->getPhasefield(), 1-i, 1-j, viscosity));
        predictorProb->addMatrixOperator(*opLaplaceUi1, i, j);
      }
    }
    
    /// < alpha*grad(u'_i) , grad(psi) >
    Operator *opLaplaceUi = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(i));
      opLaplaceUi->addSecondOrderTerm(new Phase_SOT(geometryProb->getPhasefield(), viscosity));
    predictorProb->addMatrixOperator(*opLaplaceUi, i, i);

//     Operator *opLaplaceUi_mod = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(i));
//       opLaplaceUi_mod->addFirstOrderTerm(new VectorGradient_FOT(geometryProb->getPhasefield(), new Factor<WorldVector<double> >(viscosity)));
//     predictorProb->addMatrixOperator(*opLaplaceUi_mod, i, i);
};


void NavierStokesPhase_Chorin::fillBoundaryConditions()
{ FUNCNAME("NavierStokesPhase_Chorin::fillBoundaryConditions()");
  MSG_DBG("NavierStokesPhase_Chorin::fillBoundaryConditions()\n");

  DOFVector<double> *zeroDOF = new DOFVector<double>(predictorProb->getFeSpace(0), "zero");
  zeroDOF->set(0.0);

  /// at rigid wall: no-slip boundary condition
  for (unsigned j = 0; j < dim; ++j) {
    predictorProb->addDirichletBC(LINKS, j, j, zeroDOF);
    predictorProb->addDirichletBC(RECHTS, j, j, zeroDOF);
  }
  
  /// at inflow: prescribed velocity
  for (unsigned j = 0; j < dim; ++j) {
    predictorProb->addDirichletBC(UNTEN, j, j, boundaryDOF[j]);
    //predictorProb->addDirichletBC(OBEN, j, j, boundaryDOF[j]); /// test: Dirichlet-BC at outflow boundary
  }
  
  /// at outflow: constant pressure
  pressureProb->addDirichletBC(UNTEN, 0,0, zeroDOF);
  pressureProb->addDirichletBC(OBEN, 0,0, zeroDOF);


  /// boundaryManager wird von Diagonale kopiert, wenn MatrixOperator hinzugefuegt wird -> erst MatrixOperatoren, danach erst BoundaryOperatoren
//   for (unsigned i = 0; i < dim; ++i) {
//     for (unsigned j = 0; j < dim; ++j) {
//       Operator *opLaplaceBoundary = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(j));
//       opLaplaceBoundary->addFirstOrderTerm(new ViscosityPartialDerivativeBoundary_FOT(
//         chProb.getSolution()->getDOFVector(0), i, j, viscosity1, viscosity2), GRD_PHI);
//       for (unsigned b= 0; b < nsNeumannBC.size(); ++b) {
//         if (nsNeumannBC[b] != 0) {
//           predictorProb->addBoundaryMatrixOperator(nsNeumannBC[b], opLaplaceBoundary, i, j);
//         }
//       }
//     }
//   }
};


void NavierStokesPhase_Chorin::initTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokesPhase_Chorin::initTimestep()");
  MSG_DBG("NavierStokesPhase_Chorin::initTimestep()\n");

  if (adaptInfo->getTimestepNumber() > 1) {
    WorldVector<DOFVector<double>* > gradPressure;
    for (int i = 0; i < gradPressure.getSize(); ++i)
      gradPressure[i] = new DOFVector<double>(correctorProb->getFeSpace(0), "gradPressure_i");
    DOFVector<WorldVector<double> > gradPressure_temp(correctorProb->getFeSpace(0), "gradPressure");
  
    pressureProb->getSolution()->getDOFVector(0)->getRecoveryGradient(&gradPressure_temp);
    transform(&gradPressure_temp, &gradPressure);
  
    for (int i = 0; i < gradPressure.getSize(); ++i) {
      transformDOF(predictorProb->getExactSolution(i),
        gradPressure[i],
        boundaryDOF[i], new AddFactor<double>(*getTau()));
      transformDOF(boundaryDOF[i],
        geometryProb->getPhasefield(),
        boundaryDOF[i], new Mult<double>);
    } 
    
    // delete garbage
    for(int i= 0; i<gradPressure.getSize(); ++i)
      delete gradPressure[i];
  }
};


void NavierStokesPhase_Chorin::closeTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokesPhase_Chorin::closeTimestep()");
  MSG_DBG("NavierStokesPhase_Chorin::closeTimestep()\n");

  for (int i = 0; i < correctorProb->getNumComponents(); ++i) {
    transformDOF(correctorProb->getSolution()->getDOFVector(0),
      geometryProb->getPhasefield(),
      correctorProb->getSolution()->getDOFVector(0), new Mult<double>);
  }

  int outputPeriod = 1;
  Initfile::get("user parameter->write every ith timestep", outputPeriod);
  if (adaptInfo->getTimestepNumber() % outputPeriod == 0
    && adaptInfo->getStartTime() < adaptInfo->getEndTime()) {
    writeFiles(adaptInfo, false);
  }
};

