#include "NavierStokesPhase_TaylorHood.h"

using namespace std;
using namespace AMDiS;

NavierStokesPhase_TaylorHood::NavierStokesPhase_TaylorHood(const std::string &name_) :
  super(name_),
  phase(NULL),
  bcFct(NULL),
  beta(1.0),
  epsilon(1.e-1),
  alpha(3.0),
  fileWriterPhase(NULL)
{
  for (int i = 0; i < bcDOF.getSize(); ++i)
    bcDOF[i] = NULL;
  
  Parameters::get(name + "->beta", beta);
  Parameters::get(name + "->epsilon", epsilon);
  Parameters::get(name + "->alpha", alpha);
  
};

NavierStokesPhase_TaylorHood::~NavierStokesPhase_TaylorHood()
{
  if (fileWriterPhase != NULL)
    delete fileWriterPhase;
}

void NavierStokesPhase_TaylorHood::initData()
{
  super::initData();
  phaseOld = new DOFVector<double>(getFeSpace(0), "phaseOld");

  dbc_factor = beta/pow(epsilon, alpha);
  MSG_DBG("dbc_factor = %f\n", dbc_factor);
}


void NavierStokesPhase_TaylorHood::transferInitialSolution(AdaptInfo *adaptInfo) 
{ FUNCNAME("NavierStokesPhase_TaylorHood::transferInitialSolution()");

  fileWriterPhase = new FileWriter(name + "->phase->output", getFeSpace()->getMesh(), phase);
  for (int i = 0; i < dow; i++)
    transformDOF(prob->getSolution()->getDOFVector(0), phase, prob->getSolution()->getDOFVector(0), new AMDiS::Mult<double>);

  super::transferInitialSolution(adaptInfo);
  phaseOld->interpol(phase);
}


void NavierStokesPhase_TaylorHood::closeTimestep(AdaptInfo *adaptInfo)
{
  super::closeTimestep(adaptInfo);
  
  phaseOld->interpol(phase);
  fileWriterPhase->writeFiles(adaptInfo, false);
}


void NavierStokesPhase_TaylorHood::fillOperators()
{ FUNCNAME("NavierStokesPhase_TaylorHood::fillOperators()");

  WorldVector<DOFVector<double>* > vel;
  for (size_t i = 0; i < dow; i++)
    vel[i] = prob->getSolution()->getDOFVector(i);

  TEST_EXIT(phase!=NULL && phaseOld!=NULL)("PhaseField is not set!\n");

  // fill operators for prob
  for (size_t i = 0; i < dow; ++i) {
    /// < (1/tau)*u'_i , psi >
    Operator *opTime = new Operator(prob->getFeSpace(i), prob->getFeSpace(i));
    opTime->addTerm(new Phase_ZOT(phase, density));
    prob->addMatrixOperator(*opTime, i, i, getInvTau());
    /// < (1/tau)*u_i^old , psi >
    Operator *opTimeOld = new Operator(prob->getFeSpace(i), prob->getFeSpace(i));
    opTimeOld->addTerm(new Phase_ZOT(phaseOld, density));
    opTimeOld->setUhOld(prob->getSolution()->getDOFVector(i));
    prob->addVectorOperator(*opTimeOld, i, getInvTau());
 
    /// < u^old*grad(u_i^old) , psi >
    Operator *opUGradU0 = new Operator(prob->getFeSpace(i),prob->getFeSpace(i));
    opUGradU0->addTerm(new WorldVecPhase_FOT(phase,vel, -density), GRD_PHI);
    opUGradU0->setUhOld(prob->getSolution()->getDOFVector(i));
    if (nonLinTerm == 0) {
      prob->addVectorOperator(*opUGradU0, i);
    } else {
      prob->addVectorOperator(*opUGradU0, i, &theta1);
    }

    if (nonLinTerm == 1) {
      /// < u'*grad(u_i^old) , psi >
      for (unsigned j = 0; j < dow; ++j) {
        Operator *opUGradU1 = new Operator(prob->getFeSpace(i),prob->getFeSpace(i));
        opUGradU1->addTerm(new VecAndPartialDerivative_ZOT(
          phase,
          prob->getSolution()->getDOFVector(i), j, density));
        prob->addMatrixOperator(*opUGradU1, i, j, &theta);
      }
    } else if (nonLinTerm == 2) {
      /// < u^old*grad(u'_i) , psi >
      for(unsigned j = 0; j < dow; ++j) {
        Operator *opUGradU2 = new Operator(prob->getFeSpace(i),prob->getFeSpace(i));
        opUGradU2->addTerm(new Vec2ProductPartial_FOT(
          phase,
          prob->getSolution()->getDOFVector(j), j, density), GRD_PHI);
        prob->addMatrixOperator(*opUGradU2, i, i, &theta);
      }
    }

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(prob->getFeSpace(i),prob->getFeSpace(dow));
    opGradP->addTerm(new VecAndPartialDerivative_FOT(phase,i, -1.0), GRD_PSI);
    prob->addMatrixOperator(*opGradP, i, dow);
    Operator *opGradP2 = new Operator(prob->getFeSpace(i),prob->getFeSpace(dow));
    opGradP2->addTerm(new PartialDerivative_ZOT(phase,i, -1.0));
    prob->addMatrixOperator(*opGradP2, i, dow);
  
    /// external force, i.e. gravitational force
    if (norm(force) > DBL_TOL) {
      Operator *opForce = new Operator(prob->getFeSpace(i), prob->getFeSpace(i));
      opForce->addTerm(new Phase_ZOT(phase,force[i]));
      prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = grad(phi)*g
  for (size_t i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(prob->getFeSpace(dow),prob->getFeSpace(i));
    opDivU->addTerm(new VecAndPartialDerivative_FOT(phase,i), GRD_PHI);
    prob->addMatrixOperator(*opDivU, dow, i);
    Operator *opDivU2 = new Operator(prob->getFeSpace(dow),prob->getFeSpace(i));
    opDivU2->addTerm(new PartialDerivative_ZOT(phase,i));
    prob->addMatrixOperator(*opDivU2, dow, i);
  }
}


void NavierStokesPhase_TaylorHood::addLaplaceTerm(int i)
{ FUNCNAME("NavierStokesPhase_TaylorHood::addLaplaceTerm()");

    /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
    if (laplaceType == 1) {
      for (unsigned j = 0; j < dow; ++j) {
        Operator *opLaplaceUi1 = new Operator(prob->getFeSpace(i), prob->getFeSpace(j));
        opLaplaceUi1->addTerm(new MatrixIJPhase_SOT(phase, 1-i, 1-j, viscosity));
        prob->addMatrixOperator(*opLaplaceUi1, i, j, &theta);

        opLaplaceUi1->setUhOld(prob->getSolution()->getDOFVector(j));
        prob->addVectorOperator(*opLaplaceUi1, i, &minusTheta1);
      }
    }
    
    /// < alpha*grad(u'_i) , grad(psi) >
    Operator *opLaplaceUi = new Operator(prob->getFeSpace(i), prob->getFeSpace(i));
      opLaplaceUi->addTerm(new Phase_SOT(phase, viscosity));
    prob->addMatrixOperator(*opLaplaceUi, i, i, &theta);

    opLaplaceUi->setUhOld(prob->getSolution()->getDOFVector(i));
    prob->addVectorOperator(*opLaplaceUi, i, &minusTheta1);
}


void NavierStokesPhase_TaylorHood::fillBoundaryConditions()
{ FUNCNAME("NavierStokesPhase_TaylorHood::fillBoundaryConditions()");

  for (size_t i = 0; i < dow; ++i) {
    /// dirichlet boundary condition
    /// beta/eps^alpha (1-phase)*(u_i - g_i)
    Operator *opDbcPhaseInvU = new Operator(prob->getFeSpace(i), prob->getFeSpace(i));
    opDbcPhaseInvU->addTerm(new Simple_ZOT(dbc_factor));
    opDbcPhaseInvU->addTerm(new Phase_ZOT(phase, -dbc_factor));
    prob->addMatrixOperator(*opDbcPhaseInvU, i, i);
    
    if (bcDOF[i] != NULL) {
      opDbcPhaseInvU->setUhOld(bcDOF[i]);
      prob->addVectorOperator(*opDbcPhaseInvU, i);
    } else if (bcFct != NULL) {
      Operator *opDbcPhaseInvG = new Operator(prob->getFeSpace(i));
      opDbcPhaseInvG->addTerm(new CoordsAtQP_ZOT(new ComponentExtractor(bcFct, i, dbc_factor)));
      opDbcPhaseInvG->addTerm(new VecAndCoordsAtQP_ZOT(phase, new VecAndComponentExtractor(bcFct, i, -dbc_factor)));
      prob->addVectorOperator(*opDbcPhaseInvG, i);
    }
  }
  
  /// < grad(phi)*G , psi >
  if (bcDOF[0] != NULL) {
    Operator *opGradPhiG = new Operator(prob->getFeSpace(dow),prob->getFeSpace(0));
    opGradPhiG->addTerm(new WorldVec_FOT(bcDOF), GRD_PHI);
    opGradPhiG->setUhOld(phase);
    prob->addVectorOperator(*opGradPhiG, dow);
  }
    
  /// < grad(phi)*G(x) , psi >
  if (bcFct != NULL) {
    Operator *opGradPhiG = new Operator(prob->getFeSpace(dow),prob->getFeSpace(0));
    opGradPhiG->addTerm(new VecFctAtQP_FOT(bcFct), GRD_PHI);
    opGradPhiG->setUhOld(phase);
    prob->addVectorOperator(*opGradPhiG, dow);
  }
  
  // test for implicit boundary condition
  if (bcFct == NULL) {
    bool bcCond = false;
    for (size_t i = 0; i < dow; i++)
      bcCond = bcCond || bcDOF[i] != NULL;
    TEST_EXIT(bcCond)("no implicit boundary condition specified!\n");
  } else {
    bool bcCond = false;
    for (size_t i = 0; i < dow; i++)
      bcCond = bcCond || bcDOF[i] != NULL;
    TEST_EXIT(!bcCond)("implicit boundary condition specified twice, by bcFct and bcDOF. This is not allowed!\n");
  }
}
