/** \file NavierStokesPhase_TaylorHood.h */

#ifndef NAVIER_STOKES_PHASE_TAYLOR_HOOD_H
#define NAVIER_STOKES_PHASE_TAYLOR_HOOD_H

#include "NavierStokes_TaylorHood.h"

using namespace AMDiS;

/** \ingroup NavierStokes_TaylorHood
 * \brief
 * Navier-Stokes problem in diffuse domain, using Taylor Hood elements
 * 
 * dt(phase*u_j) + phase*u*grad(u_j) = phase*d_j(p) - nu*div(phase*grad(u_j)) - beta/eps^alpha (1-phase)*(u_j - g_j); j=1...d
 * div(phase*u) = grad(phase)*g
 */
class NavierStokesPhase_TaylorHood : public NavierStokes_TaylorHood
{
public: // typedefs

  typedef NavierStokes_TaylorHood super;

public: // methods

  NavierStokesPhase_TaylorHood(const std::string &name_);
  ~NavierStokesPhase_TaylorHood();
  
  virtual void initData();
  virtual void transferInitialSolution(AdaptInfo *adaptInfo);  
  virtual void closeTimestep(AdaptInfo *adaptInfo);
  
  // === getting/setting methods ===
  
  DOFVector<double> *getPhase()
  {
    return phase;
  }

  DOFVector<double> *getBcDOF(int i)
  {
    return bcDOF[i];
  }
  
  AbstractFunction<WorldVector<double>, WorldVector<double> >* getBcFct()
  {
    return bcFct;
  }
  
  double getEpsilon()
  {
    return epsilon;
  }
  
  void setPhase(DOFVector<double>* phase_)
  {
    phase = phase_;
  }
  
  void setBcDOF(DOFVector<double>* bcDOF_i, int i)
  {
    bcDOF[i] = bcDOF_i;
  }
  
  void setBcFct(AbstractFunction<WorldVector<double>, WorldVector<double> >* fct_)
  {
    bcFct = fct_;
  }

  void setEpsilon(double epsilon_)
  {
    epsilon = epsilon_;
  }

// protected: // methods

  virtual void fillOperators();
  virtual void fillBoundaryConditions();
  virtual void addLaplaceTerm(int i);

private: // variables
  
  double dbc_factor; // beta/epsilon^alpha
  double beta;
  double epsilon;
  double alpha;

  DOFVector<double> *phase;
  DOFVector<double> *phaseOld;

  WorldVector<DOFVector<double>*> bcDOF;
  AbstractFunction<WorldVector<double>, WorldVector<double> >* bcFct;
  
  FileWriter *fileWriterPhase;
};


struct ComponentExtractor : AbstractFunction<double, WorldVector<double> >
{
  ComponentExtractor(AbstractFunction<WorldVector<double>, WorldVector<double> >* fct_, int comp_, double factor_=1.0)
  : AbstractFunction<double, WorldVector<double> >(fct_->getDegree()),
    fct(fct_), 
    comp(comp_),
    factor(factor_) {}
  
  double operator()(const WorldVector<double>& x) const {
    WorldVector<double> result = (*fct)(x);
    return factor*result[comp];
  }
  
private:
  AbstractFunction<WorldVector<double>, WorldVector<double> >* fct;
  int comp;
  double factor;
};


struct VecAndComponentExtractor : BinaryAbstractFunction<double, double, WorldVector<double> >
{
  VecAndComponentExtractor(AbstractFunction<WorldVector<double>, WorldVector<double> >* fct_, int comp_, double factor_=1.0)
  : BinaryAbstractFunction<double, double, WorldVector<double> >(fct_->getDegree()+1),
    fct(fct_), 
    comp(comp_),
    factor(factor_) {}
  
  double operator()(const double& vec, const WorldVector<double>& x) const {
    WorldVector<double> result = (*fct)(x);
    return factor*vec*result[comp];
  }
  
private:
  AbstractFunction<WorldVector<double>, WorldVector<double> >* fct;
  int comp;
  double factor;
};

#endif // NAVIER_STOKES_PHASE_TAYLOR_HOOD_H
