/** \file NavierStokes_TaylorHood.h */

#ifndef NAVIER_STOKES_TAYLOR_HOOD_H
#define NAVIER_STOKES_TAYLOR_HOOD_H

#include "AMDiS.h"
#include "Helpers.h"
#include "POperators.h"
#include "BoundaryFunctions.h"
#include "BaseProblem.h"
#include "ExtendedProblemStat.h"
#include <boost/numeric/mtl/mtl.hpp>
#include <boost/numeric/mtl/utility/property_map.hpp>

// boundary numbers
#define LINKS 1
#define RECHTS 2
#define UNTEN 3
#define OBEN 4

using namespace AMDiS;

/** \ingroup NavierStokes_TaylorHood
 * \brief
 * Navier-Stokes problem, using Taylor Hood elements
 */
class NavierStokes_TaylorHood : public BaseProblem<ExtendedProblemStat>
{
public: // typedefs
  
  typedef BaseProblem<ExtendedProblemStat> super;
  
public: // methods

  NavierStokes_TaylorHood(const std::string &name_);

  /// if initialVelocityIsSet is true delete initialX and initialY
  ~NavierStokes_TaylorHood();

  virtual void initData();
  virtual void solveInitialProblem(AdaptInfo *adaptInfo);

  virtual void transferInitialSolution(AdaptInfo *adaptInfo);  

  virtual void closeTimestep(AdaptInfo *adaptInfo);

  // === getting/setting methods ===

  /// set dimension of the macro-mesh
  void setDimension(WorldVector<double> dimension_) { dimension = dimension_; }

  DOFVector<WorldVector<double> >* getVelocity()
  {
    if (velocity == NULL)
      velocity = new DOFVector<WorldVector<double> >(getFeSpace(0), "velocity");
    return velocity;
  }

// protected: // methods

  virtual void fillOperators();
  virtual void fillBoundaryConditions();
  
  virtual void addLaplaceTerm(int i);

protected: // variables

  AbstractFunction<double, WorldVector<double> > *initialX, *initialY;
  WorldVector<DOFVector<double>*> initialVel;
  std::vector<WorldVector<double> > dirichletPoints;

  DOFVector<WorldVector<double> >* velocity;

  void calcVelocity()
  {
    if (dow == 1)
      transformDOF(prob->getSolution()->getDOFVector(0), velocity, new AMDiS::Vec1WorldVec<double>);
    else if (dow == 2)
      transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), velocity, new AMDiS::Vec2WorldVec<double>);
    else if (dow == 3)
      transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), prob->getSolution()->getDOFVector(2), velocity, new AMDiS::Vec3WorldVec<double>);
  }

  bool forceDBC;
  bool initialVelocityIsSet;

  int laplaceType;
  int nonLinTerm;

  double oldTimestep;
  double viscosity;
  double density;
  double c;
  double theta;
  double theta1;
  double minusTheta1;

  WorldVector<double> force;
  WorldVector<double> dimension;

  FileVectorWriter *fileWriter;
};
#endif // NAVIER_STOKES_TAYLOR_HOOD_H
