/** \file PhaseFieldCrystal.h */

#ifndef PHASE_FIELD_CRYSTAL_H
#define PHASE_FIELD_CRYSTAL_H

#include "AMDiS.h"
#include "BaseProblem.h"

using namespace AMDiS;

/** Phase-field Crystal problem
 */
class PhaseFieldCrystal : public StandardBaseProblem
{
public: // typedefs

  typedef StandardBaseProblem super;

public:

  PhaseFieldCrystal(const std::string &name_);
  ~PhaseFieldCrystal() {};

  double *getTempParameter() { return &tempParameter; }

  virtual void fillOperators();
  virtual void fillBoundaryConditions();

protected:

  bool useMobility;

  double tempParameter;
  double r;
  double rho0;
  double density;
  double two;
  double minus2;
};


/** \ingroup MainInstat
 * \brief
 * Abstract function to calculate the pure PFC-Energy
 */
class Energy : public BinaryAbstractFunction<double,double,double>
{
  public:
    Energy() : BinaryAbstractFunction<double,double,double>(4) { }

    double operator()(const double &rho, const double &mu) const {
      return -0.25*sqr(sqr(rho)) + 0.5*rho*mu; }
};


class MobilityPfc : public AbstractFunction<double,double>
{
  public:
    MobilityPfc(double density_ = -0.3, double factor_ = 1.0) : 
      AbstractFunction<double,double>(1),
      density(density_),
      factor(factor_),
      delta(1.e-6) { }

    double operator()(const double &rho) const 
    {
      double mobility= abs(rho - 3.0*density)*factor;
      return std::max(mobility, delta);
    }
  
  protected:
    double density;
    double factor;
    double delta;
};

class Pow2Functor : public AbstractFunction<double,double>
{
  public:
    Pow2Functor(double factor_=1.0) :
      AbstractFunction<double,double>(2),
      factor(factor_) { }
    double operator()(const double &ch) const
    {
      return factor * sqr(ch);
    }

  protected:
    double factor;
};

class Pow3Functor : public AbstractFunction<double,double>
{
  public:
    Pow3Functor(double factor_=1.0) :
      AbstractFunction<double,double>(3),
      factor(factor_) { }
    double operator()(const double &ch) const
    {
      return factor * sqr(ch) * ch;
    }

  protected:
    double factor;
};

#endif // PHASE_FIELD_CRYSTAL_H
