/** \file chns.h */

#ifndef CHNS_FUNCTORS_H
#define CHNS_FUNCTORS_H

#include "AMDiS.h"

/** \brief
 * Abstract function for Cahn-Hilliard mobility
 */
class MobilityCH0 : public AbstractFunction<double,double>
{
  public:
    MobilityCH0(double gamma_=1.0) : 
      AbstractFunction<double,double>(4), 
      gamma(gamma_),
      delta(1.e-6) { }
    double operator()(const double &ch) const 
    {
      double phase = std::max(0.0, std::min(1.0, ch));
      double mobility = 0.25*sqr(phase)*sqr(phase-1.0);
      return gamma * std::max(mobility, delta);
    }

  protected:
    double gamma;
    double delta;
};

class MobilityCH1 : public AbstractFunction<double,double>
{
  public:
    MobilityCH1(double gamma_=1.0) :
      AbstractFunction<double,double>(4),
      gamma(gamma_),
      delta(1.e-6) { }
    double operator()(const double &ch) const
    {
      double phase = std::max(-1.0, std::min(1.0, ch));
      double mobility = 0.25*sqr(sqr(phase)-1.0);
      return gamma * std::max(mobility, delta);
    }

  protected:
    double gamma;
    double delta;
};

class ScaleFunctor : public AbstractFunction<double,double>
{
  public:
    ScaleFunctor(double factor_=1.0) :
      AbstractFunction<double,double>(1),
      factor(factor_) { }
    double operator()(const double &ch) const
    {
      return factor * ch;
    }

  protected:
    double factor;
};

class Pow2Functor : public AbstractFunction<double,double>
{
  public:
    Pow2Functor(double factor_=1.0) :
      AbstractFunction<double,double>(2),
      factor(factor_) { }
    double operator()(const double &ch) const
    {
      return factor * sqr(ch);
    }

  protected:
    double factor;
};

class Pow3Functor : public AbstractFunction<double,double>
{
  public:
    Pow3Functor(double factor_=1.0) :
      AbstractFunction<double,double>(3),
      factor(factor_) { }
    double operator()(const double &ch) const
    {
      return factor * sqr(ch) * ch;
    }

  protected:
    double factor;
};


class MobilityCH0Diff : public BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >
{
  public:
    MobilityCH0Diff(double factor_=1.0) :
      BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >(4),
      factor(factor_) { }
    WorldVector<double> operator()(const double &c, const WorldVector<double>& grdMu) const
    {
      double phase = std::max(0.0, std::min(1.0, c));
      return (factor * (sqr(phase)*phase - 1.5*sqr(phase) + 0.5*phase)) * grdMu;
    }

  protected:
    double factor;
};

class MobilityCH1Diff : public BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >
{
  public:
    MobilityCH1Diff(double factor_=1.0) :
      BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >(4),
      factor(factor_) { }
    WorldVector<double> operator()(const double &c, const WorldVector<double>& grdMu) const
    {
      double phase = std::max(-1.0, std::min(1.0, c));
      return (factor * (sqr(phase)*phase - phase)) * grdMu;
    }

  protected:
    double factor;
};

class DoubleWell0 : public AbstractFunction<double,double>
{
  public:
    DoubleWell0(double factor_=1.0) : 
      AbstractFunction<double,double>(4), 
      factor(factor_) { }
      
    double operator()(const double &c) const 
    {
      return factor * (sqr(c)*c - 1.5*sqr(c) + 0.5*c);
    }

  protected:
    double factor;
};

class DoubleWell1 : public AbstractFunction<double,double>
{
  public:
    DoubleWell1(double factor_=1.0) : 
      AbstractFunction<double,double>(4), 
      factor(factor_) { }
      
    double operator()(const double &c) const
    {
      return factor * (sqr(c)*c - c);
    }

  protected:
    double factor;
};

class DoubleWell0Diff : public AbstractFunction<double,double>
{
  public:
    DoubleWell0Diff(double factor_=1.0) : 
      AbstractFunction<double,double>(4), 
      factor(factor_) { }
      
    double operator()(const double &c) const 
    {
      return factor * (3.0*sqr(c) - 3.0*c + 0.5);
    }

  protected:
    double factor;
};

class DoubleWell1Diff : public AbstractFunction<double,double>
{
  public:
    DoubleWell1Diff(double factor_=1.0) : 
      AbstractFunction<double,double>(4), 
      factor(factor_) { }
      
    double operator()(const double &c) const
    {
      return factor * (3.0*sqr(c) - 1.0);
    }

  protected:
    double factor;
};

struct LinearInterpolation : AbstractFunction<double,double>
{
  LinearInterpolation(double v0_, double v1_, double factor_ = 1.0) : 
	AbstractFunction<double,double>(4), 
	v0(v0_), v1(v1_), factor(factor_) { }
      
  double operator()(const double &c) const
  {
	return factor * (v0*(c+1.0) - v1*(c-1.0))/2.0;
  }

private:
  double v0;
  double v1;
  double factor;
};

struct LinearInterpolationFactor : BinaryAbstractFunction<double, double, double>
{
  LinearInterpolationFactor(double v0_, double v1_, double factor_ = 1.0) : 
	BinaryAbstractFunction<double, double, double>(4), 
	v0(v0_), v1(v1_), factor(factor_) { }
      
  double operator()(const double &c, const double& u) const
  {
	return factor * u * (v0*(c+1.0) - v1*(c-1.0))/2.0;
  }

private:
  double v0;
  double v1;
  double factor;
};

struct LinearInterpolationDiffFactor : BinaryAbstractFunction<double,double,double>
{
  LinearInterpolationDiffFactor(double v0_, double v1_, double factor_ = 1.0) : 
	BinaryAbstractFunction<double,double,double>(4), 
	v0(v0_), v1(v1_), factor(factor_) { }
      
  double operator()(const double &c, const double& z) const
  {
	return factor * z * (v0 - v1)/2.0;
  }

private:
  double v0;
  double v1;
  double factor;
};

struct LinearInterpolationDiffVector : BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >
{
  LinearInterpolationDiffVector(double v0_, double v1_, double factor_ = 1.0) : 
	BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >(4), 
	v0(v0_), v1(v1_), factor(factor_) { }
      
  WorldVector<double> operator()(const double &c, const WorldVector<double>& grdU) const
  {
	return (factor * (v0 - v1)/2.0) * grdU;
  }

private:
  double v0;
  double v1;
  double factor;
};

struct TaylorExpansionFactor : BinaryAbstractFunction<double,double,double>
{
  TaylorExpansionFactor(double v0_, double v1_, double factor_ = 1.0) : 
	BinaryAbstractFunction<double,double,double>(4), 
	v0(v0_), v1(v1_), factor(factor_) { }
      
  double operator()(const double &c, const double& u) const
  {
	return LinearInterpolationFactor(v0, v1, factor)(c, u) + c * LinearInterpolationDiffFactor(v0, v1, factor)(c, u);
  }

private:
  double v0;
  double v1;
  double factor;
};

#endif // CHNS_FUNCTORS_H
