/***********************************************************************
 * Software License Agreement (BSD License)
 *
 * Copyright 2011 Jose Luis Blanco (joseluisblancoc@gmail.com).
 *   All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *************************************************************************/
/** \file kdtree_nanoflann_mesh.h */

#ifndef KDTREE_NANOFLANN_MESH_H
#define KDTREE_NANOFLANN_MESH_H

#include <nanoflann.hpp>

#include <cstdlib>
#include <iostream>

#include "AMDiS.h"
#include "ElementFunction.h"
#include "GeometryTools.h"

using namespace nanoflann;
using namespace AMDiS;

namespace experimental {

// =====================================================================================
typedef WorldVector<double> PointType;
typedef boost::tuple<const MacroElement*, int, unsigned long> DataType;
typedef std::vector<PointType> VectorOfPointsType;
typedef std::vector<DataType> VectorOfDataType;
// =====================================================================================


  /** A simple vector-of-vectors adaptor for nanoflann, without duplicating the storage.
    *  The i'th vector represents a point in the state space.
    *
    *  \tparam DIM If set to >0, it specifies a compile-time fixed dimensionality for the points in the data set, allowing more compiler optimizations.
    *  \tparam value_type The type of the point coordinates (typically, double or float).
    *  \tparam Distance The distance metric to use: nanoflann::metric_L1, nanoflann::metric_L2, nanoflann::metric_L2_Simple, etc.
    *  \tparam index_type The type for indices in the KD-tree index (typically, size_t of int)
    */
  template < class VectorOfPointsType_, 
	     class VectorOfDataType_,
	     typename value_type = double, 
	     int DIM = -1, 
	     class Distance = nanoflann::metric_L2_Simple, 
	     typename index_type = DegreeOfFreedom >
  struct KDTreeMeshAdaptor
  {
    typedef KDTreeMeshAdaptor<  VectorOfPointsType_,
				VectorOfDataType_,
				value_type, 
				DIM, 
				Distance,
				index_type > self_t;
    typedef typename Distance::template traits<value_type, self_t>::distance_t metric_t;
    typedef KDTreeSingleIndexAdaptor< metric_t, 
				      self_t, 
				      DIM, 
				      index_type > index_t;

    index_t* index; //! The kd-tree index for the user to call its methods as usual with any other FLANN index.
    Mesh* mesh;

    /// Constructor: takes a const ref to the vector of vectors object with the data points
    KDTreeMeshAdaptor(const int dimensionality,
		      Mesh* mesh_,
		      const int leaf_max_size = 10)
    : mesh(mesh_)
    {
      const size_t dims = Global::getGeo(WORLD);
      if (mesh->getNumberOfLeaves() == 0)
	mesh->updateNumberOfLeaves();
      
      assert(mesh->getNumberOfLeaves() != 0 && dims != 0);
      if (DIM > 0 && static_cast<int>(dims) != DIM)
	throw std::runtime_error("Point dimensionality does not match the 'DIM' template argument");
      
      init();
      
      index = new index_t( dims, *this /* adaptor */, nanoflann::KDTreeSingleIndexAdaptorParams(leaf_max_size, dims ) );
      index->buildIndex();
    }

    ~KDTreeMeshAdaptor() {
      delete index;
    }
    
    void init()
    {
      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS);

      int dim = mesh->getDim();
      while (elInfo) {
	ElInfo* macroElInfo = mesh->createNewElInfo();
	macroElInfo->fillMacroInfo(elInfo->getMacroElement());
	stack.fillRefinementPath(*elInfo, *macroElInfo);

	DataType elInfoData = boost::tuples::make_tuple(elInfo->getMacroElement(), elInfo->getRefinementPathLength(), elInfo->getRefinementPath());

	WorldVector<double> centroid = elInfo->getCoord(0);
	for (int i = 1; i < dim+1; i++)
	  centroid += elInfo->getCoord(i);
	centroid *= 1.0/(dim+1.0);

	m_points.push_back(centroid);
	m_data.push_back(elInfoData);

	elInfo = stack.traverseNext(elInfo);
      }
    }
    
    void reinit(const int leaf_max_size = 10)
    {
      delete index;
      const size_t dims = Global::getGeo(WORLD);
      m_points.clear();
      m_data.clear();
      init();
      index = new index_t( dims, *this /* adaptor */, nanoflann::KDTreeSingleIndexAdaptorParams(leaf_max_size, dims ) );
      index->buildIndex();
    }

    VectorOfPointsType_ m_points;
    VectorOfDataType_ m_data;

    /** Query for the \a num_closest closest points to a given point (entered as query_point[0:dim-1]).
      *  Note that this is a short-cut method for index->findNeighbors().
      *  The user can also call index->... methods as desired.
      * \note nChecks_IGNORED is ignored but kept for compatibility with the original FLANN interface.
      */
    inline void query(const value_type *query_point, 
		      const size_t num_closest, 
		      index_type *out_indices, 
		      value_type *out_distances_sq, 
		      const int nChecks_IGNORED = 10) const
    {
      nanoflann::KNNResultSet<value_type, index_type> resultSet(num_closest);
      resultSet.init(out_indices, out_distances_sq);
      index->findNeighbors(resultSet, query_point, nanoflann::SearchParams());
    }

    /** @name Interface expected by KDTreeSingleIndexAdaptor
      * @{ */

    const self_t & derived() const {
      return *this;
    }
    self_t & derived() {
      return *this;
    }

    // Must return the number of data points
    inline size_t kdtree_get_point_count() const {
      return m_points.size();
    }

    // Returns the distance between the vector "p1[0:size-1]" and the data point with index "idx_p2" stored in the class:
    inline value_type kdtree_distance(const value_type *p1, 
				      const index_type idx_p2, 
				      size_t size) const
    {
      value_type s = 0.0;
      for (size_t i = 0; i < size; i++) {
	const value_type d = p1[i] - m_points[idx_p2][i];
	s += d*d;
      }
      return s;
    }

    // Returns the dim'th component of the idx'th point in the class:
    inline value_type kdtree_get_pt(const index_type idx, size_t dim) const {
      return m_points[idx][dim];
    }

    // Optional bounding-box computation: return false to default to a standard bbox computation loop.
    //   Return true if the BBOX was already computed by the class and returned in "bb" so it can be avoided to redo it again.
    //   Look at bb.size() to find out the expected dimensionality (e.g. 2 or 3 for point clouds)
    template <class BBOX>
    bool kdtree_get_bbox(BBOX &bb) const {
      return false;
    }

    /** @} */
    
    
    /**
    * strategy and nrOfPoints are ignored
    *  == evalAtPoint_simple
    **/
    template<typename T>
    T evalAtPoint(DOFVector<T>& vec, PointType& x, int strategy = 0, int nrOfPoints = -1)
    {
      ElInfo* elInfo = vec.getFeSpace()->getMesh()->createNewElInfo();
      elInfo->setFillFlag(Mesh::FILL_COORDS);
      T value; nullify(value);
      if (getNearestElInfo(x, elInfo)) {
	int dim = vec.getFeSpace()->getMesh()->getDim();
	DimVec<double> lambda(dim, NO_INIT);

	if (dim == 1) {
	  double length = abs(elInfo->getCoord(1)[0] - elInfo->getCoord(0)[0]);
	  for (int i = 0; i < dim+1; i++)
	    lambda[i] = abs(x[0] - elInfo->getCoord((i+1)%2)[0])/length;
	}
	else if (dim == 2) {
	  double area = meshconv::triangle_area_3d(elInfo->getCoord(0).begin(), 
			  elInfo->getCoord(1).begin(), elInfo->getCoord(2).begin());
	  for (int i = 0; i < dim+1; i++)
	    lambda[i] = meshconv::triangle_area_3d(x.begin(), 
			  elInfo->getCoord((i+1)%3).begin(), elInfo->getCoord((i+2)%3).begin())/area;
	}
	else if (dim == 3) {
	  double volume = meshconv::volume_tetrahedron(elInfo->getCoord(0).begin(), 
			    elInfo->getCoord(1).begin(), elInfo->getCoord(2).begin(), elInfo->getCoord(3).begin());
	  for (int i = 0; i < dim+1; i++)
	    lambda[i] = meshconv::volume_tetrahedron(x.begin(), 
			  elInfo->getCoord((i+1)%3).begin(), elInfo->getCoord((i+2)%3).begin(), elInfo->getCoord((i+3)%3).begin())/volume;
	}

	ElementFunctionDOFVec<T> elFct(&vec);
	elFct.setElInfo(elInfo);
	value = elFct(lambda);
      }
      return value;
    }
    
    
    bool getNearestElInfo(PointType &x, ElInfo*& elInfo)
    {
      const size_t nnp = 1;
      std::vector<DegreeOfFreedom> ret_indexes(nnp);
      std::vector<double> out_dists_sqr(nnp);
      nanoflann::KNNResultSet<double, DegreeOfFreedom> resultSet(nnp);
      
      resultSet.init(&ret_indexes[0], &out_dists_sqr[0]);
      index->findNeighbors(resultSet, x.begin(), nanoflann::SearchParams());
      
      DataType data = m_data[ret_indexes[0]];
      elInfo->fillElInfo(data.get<0>(), data.get<1>(), data.get<2>());
      
      return true;
    }

  }; // end of KDTreeMeshAdaptor

  typedef KDTreeMeshAdaptor< VectorOfPointsType, VectorOfDataType, double >  KD_Tree_M;
  static std::map<const FiniteElemSpace*, std::pair<int, KD_Tree_M*> > kdtreeMap_M;

  inline KD_Tree_M* provideKDTree(const FiniteElemSpace* feSpace)
  {
    if (kdtreeMap_M.find(feSpace) != kdtreeMap_M.end()) {
      if (kdtreeMap_M[feSpace].first != feSpace->getMesh()->getChangeIndex()) {
	kdtreeMap_M[feSpace].second->reinit();
	kdtreeMap_M[feSpace].first = feSpace->getMesh()->getChangeIndex();
      }
    } else {
      KD_Tree_M* kdtree = new KD_Tree_M(Global::getGeo(WORLD), feSpace->getMesh());
      kdtreeMap_M[feSpace] = std::make_pair(feSpace->getMesh()->getChangeIndex(), kdtree);
    }

    return kdtreeMap_M[feSpace].second;
  }
  
} // end namespace experimental

#endif
